@@ -49,6 +49,7 @@ type AgentConfiguration struct {
 	HealthCheckAddr              string
 	DisconnectAfterJob           bool
 	DisconnectAfterIdleTimeout   int
+	DisconnectAfterUptime        int
 	CancelGracePeriod            int
 	SignalGracePeriod            time.Duration
 	EnableJobLogTmpfile          bool
@@ -61,6 +62,7 @@ type AgentConfiguration struct {
 	AcquireJob                   string
 	TracingBackend               string
 	TracingServiceName           string
+	TracingPropagateTraceparent  bool
 	TraceContextEncoding         string
 	DisableWarningsFor           []string
 	AllowMultipartArtifactUpload bool
