@@ -5,9 +5,11 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
+	"maps"
 	"math/rand/v2"
 	"os"
 	"path/filepath"
+	"slices"
 	"strconv"
 	"strings"
 	"sync"
@@ -111,11 +113,6 @@ type JobRunnerConfig struct {
 	AgentStdout io.Writer
 }
 
-type jobRunner interface {
-	Run(ctx context.Context) error
-	CancelAndStop() error
-}
-
 type JobRunner struct {
 	// The configuration for the job runner
 	conf JobRunnerConfig
@@ -133,7 +130,7 @@ type JobRunner struct {
 	client *core.Client
 
 	// The internal process of the job
-	process jobAPI
+	process jobProcess
 
 	// The internal buffer of the process output
 	output *process.Buffer
@@ -164,7 +161,8 @@ type JobRunner struct {
 	envJSONFile  *os.File
 }
 
-type jobAPI interface {
+// jobProcess is either a *process.Process, or a *kubernetes.Runner.
+type jobProcess interface {
 	Done() <-chan struct{}
 	Started() <-chan struct{}
 	Interrupt() error
@@ -173,14 +171,21 @@ type jobAPI interface {
 	WaitStatus() process.WaitStatus
 }
 
-var _ jobRunner = (*JobRunner)(nil)
-
 // Initializes the job runner
-func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, conf JobRunnerConfig) (jobRunner, error) {
+func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, conf JobRunnerConfig) (*JobRunner, error) {
+	// If the accept response has a token attached, we should use that instead of the Agent Access Token that
+	// our current apiClient is using
+	if conf.Job.Token != "" {
+		clientConf := apiClient.Config()
+		clientConf.Token = conf.Job.Token
+		apiClient = apiClient.New(clientConf)
+	}
+
 	r := &JobRunner{
 		agentLogger: l,
 		conf:        conf,
 		apiClient:   apiClient,
+		client:      &core.Client{APIClient: apiClient, Logger: l},
 	}
 
 	var err error
@@ -193,15 +198,6 @@ func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, con
 		conf.JobStatusInterval = 1 * time.Second
 	}
 
-	// If the accept response has a token attached, we should use that instead of the Agent Access Token that
-	// our current apiClient is using
-	if r.conf.Job.Token != "" {
-		clientConf := r.apiClient.Config()
-		clientConf.Token = r.conf.Job.Token
-		r.apiClient = api.NewClient(r.agentLogger, clientConf)
-	}
-	r.client = &core.Client{APIClient: r.apiClient, Logger: l}
-
 	// Create our header times struct
 	r.headerTimesStreamer = newHeaderTimesStreamer(r.agentLogger, r.onUploadHeaderTime)
 
@@ -222,8 +218,8 @@ func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, con
 	// TempDir is not guaranteed to exist
 	tempDir := os.TempDir()
 	if _, err := os.Stat(tempDir); os.IsNotExist(err) {
-		// Actual file permissions will be reduced by umask, and won't be 0777 unless the user has manually changed the umask to 000
-		if err = os.MkdirAll(tempDir, 0777); err != nil {
+		// Actual file permissions will be reduced by umask, and won't be 0o777 unless the user has manually changed the umask to 000
+		if err = os.MkdirAll(tempDir, 0o777); err != nil {
 			return nil, err
 		}
 	}
@@ -269,7 +265,15 @@ func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, con
 		if err != nil {
 			return nil, err
 		}
-		os.Setenv("BUILDKITE_JOB_LOG_TMPFILE", tmpFile.Name())
+
+		err := os.Chmod(tmpFile.Name(), 0o644) // Make it world-readable - useful for log collection etc
+		if err != nil {
+			return nil, fmt.Errorf("failed to set permissions on job log tmpfile %s: %w", tmpFile.Name(), err)
+		}
+
+		if err := os.Setenv("BUILDKITE_JOB_LOG_TMPFILE", tmpFile.Name()); err != nil {
+			return nil, fmt.Errorf("failed to set BUILDKITE_JOB_LOG_TMPFILE: %v", err)
+		}
 		outputWriter = io.MultiWriter(outputWriter, tmpFile)
 	}
 
@@ -339,6 +343,7 @@ func NewJobRunner(ctx context.Context, l logger.Logger, apiClient APIClient, con
 				logger.StringField("build_id", r.conf.Job.Env["BUILDKITE_BUILD_ID"]),
 				logger.StringField("build_number", r.conf.Job.Env["BUILDKITE_BUILD_NUMBER"]),
 				logger.StringField("job_url", fmt.Sprintf("%s#%s", r.conf.Job.Env["BUILDKITE_BUILD_URL"], r.conf.Job.ID)),
+				logger.StringField("build_url", r.conf.Job.Env["BUILDKITE_BUILD_URL"]),
 				logger.StringField("job_id", r.conf.Job.ID),
 				logger.StringField("step_key", r.conf.Job.Env["BUILDKITE_STEP_KEY"]),
 			)
@@ -426,13 +431,27 @@ func (r *JobRunner) createEnvironment(ctx context.Context) ([]string, error) {
 	// sent by Buildkite. The variables below should always take
 	// precedence.
 	env := make(map[string]string)
-	for key, value := range r.conf.Job.Env {
-		env[key] = value
-	}
+	maps.Copy(env, r.conf.Job.Env)
 
 	// The agent registration token should never make it into the job environment
 	delete(env, "BUILDKITE_AGENT_TOKEN")
 
+	// When in KubernetesExec mode, filter out the Kubernetes plugin,
+	// since it's not a real plugin. agent-stack-k8s reads it but we have no
+	// need for it. Supplying it when not using agent-stack-k8s is a mistake
+	// but not one worth preventing.
+	if pluginsJSON := env["BUILDKITE_PLUGINS"]; pluginsJSON != "" && r.conf.KubernetesExec {
+		filtered, err := removeKubernetesPlugin([]byte(pluginsJSON))
+		if err != nil {
+			r.agentLogger.Error("Invalid BUILDKITE_PLUGINS: %w", err)
+		}
+		if string(filtered) == "" {
+			delete(env, "BUILDKITE_PLUGINS")
+		} else {
+			env["BUILDKITE_PLUGINS"] = string(filtered)
+		}
+	}
+
 	// Write out the job environment to file:
 	// - envShellFile: in k="v" format, with newlines escaped
 	// - envJSONFile: as a single JSON object {"k":"v",...}, escaped appropriately for JSON.
@@ -440,7 +459,7 @@ func (r *JobRunner) createEnvironment(ctx context.Context) ([]string, error) {
 	// on the job upstream - and expose the path in another environment variable.
 	if r.envShellFile != nil {
 		for key, value := range env {
-			if _, err := r.envShellFile.WriteString(fmt.Sprintf("%s=%q\n", key, value)); err != nil {
+			if _, err := fmt.Fprintf(r.envShellFile, "%s=%q\n", key, value); err != nil {
 				return nil, err
 			}
 		}
@@ -490,6 +509,18 @@ func (r *JobRunner) createEnvironment(ctx context.Context) ([]string, error) {
 	env["BUILDKITE_AGENT_ACCESS_TOKEN"] = apiConfig.Token
 	env["BUILDKITE_NO_HTTP2"] = fmt.Sprint(apiConfig.DisableHTTP2)
 
+	// ... including any server-specified request headers, so that sub-processes such as
+	// buildkite-agent annotate etc can respect them.
+	for header, values := range r.apiClient.ServerSpecifiedRequestHeaders() {
+		k := fmt.Sprintf(
+			"BUILDKITE_REQUEST_HEADER_%s",
+			strings.ToUpper(strings.ReplaceAll(header, "-", "_")),
+		)
+		for _, v := range values {
+			env[k] = v
+		}
+	}
+
 	// Add agent environment variables
 	env["BUILDKITE_AGENT_DEBUG"] = fmt.Sprint(r.conf.Debug)
 	env["BUILDKITE_AGENT_DEBUG_HTTP"] = fmt.Sprint(r.conf.DebugHTTP)
@@ -497,15 +528,12 @@ func (r *JobRunner) createEnvironment(ctx context.Context) ([]string, error) {
 
 	// We know the BUILDKITE_BIN_PATH dir, because it's the path to the
 	// currently running file (there is only 1 binary)
+	// Note that [os.Executable] returns an absolute path.
 	exePath, err := os.Executable()
 	if err != nil {
 		return nil, err
 	}
-	dir, err := filepath.Abs(filepath.Dir(exePath))
-	if err != nil {
-		return nil, err
-	}
-	env["BUILDKITE_BIN_PATH"] = dir
+	env["BUILDKITE_BIN_PATH"] = filepath.Dir(exePath)
 
 	// Add options from the agent configuration
 	env["BUILDKITE_CONFIG_PATH"] = r.conf.AgentConfiguration.ConfigPath
@@ -591,6 +619,17 @@ func (r *JobRunner) createEnvironment(ctx context.Context) ([]string, error) {
 	if r.conf.AgentConfiguration.TracingBackend != "" {
 		env["BUILDKITE_TRACING_BACKEND"] = r.conf.AgentConfiguration.TracingBackend
 		env["BUILDKITE_TRACING_SERVICE_NAME"] = r.conf.AgentConfiguration.TracingServiceName
+
+		// Buildkite backend can provide a traceparent property on the job
+		// which can be propagated to the job tracing if OpenTelemetry is used
+		//
+		// https://www.w3.org/TR/trace-context/#traceparent-header
+		if r.conf.Job.TraceParent != "" {
+			env["BUILDKITE_TRACING_TRACEPARENT"] = r.conf.Job.TraceParent
+		}
+		if r.conf.AgentConfiguration.TracingPropagateTraceparent {
+			env["BUILDKITE_TRACING_PROPAGATE_TRACEPARENT"] = "true"
+		}
 	}
 
 	env["BUILDKITE_AGENT_DISABLE_WARNINGS_FOR"] = strings.Join(r.conf.AgentConfiguration.DisableWarningsFor, ",")
@@ -630,6 +669,18 @@ func truncateEnv(l logger.Logger, env map[string]string, key string, max int) er
 	return nil
 }
 
+func removeKubernetesPlugin(pluginsJSON []byte) ([]byte, error) {
+	var plugins []map[string]json.RawMessage
+	if err := json.Unmarshal(pluginsJSON, &plugins); err != nil {
+		return pluginsJSON, err
+	}
+	plugins = slices.DeleteFunc(plugins, func(plugin map[string]json.RawMessage) bool {
+		_, isK8sPlugin := plugin["github.com/buildkite-plugins/kubernetes-buildkite-plugin"]
+		return isK8sPlugin
+	})
+	return json.Marshal(plugins)
+}
+
 type LogWriter struct {
 	l logger.Logger
 }
@@ -754,7 +805,7 @@ func (r *JobRunner) jobCancellationChecker(ctx context.Context, wg *sync.WaitGro
 }
 
 func (r *JobRunner) onUploadHeaderTime(ctx context.Context, cursor, total int, times map[string]string) {
-	roko.NewRetrier(
+	err := roko.NewRetrier(
 		roko.WithMaxAttempts(10),
 		roko.WithStrategy(roko.Constant(5*time.Second)),
 	).DoWithContext(ctx, func(retrier *roko.Retrier) error {
@@ -770,6 +821,10 @@ func (r *JobRunner) onUploadHeaderTime(ctx context.Context, cursor, total int, t
 
 		return err
 	})
+
+	if err != nil {
+		r.agentLogger.Error("Ultimately unable to upload header times: %v", err)
+	}
 }
 
 // jobLogger is just a simple wrapper around a JSON Logger that satisfies the
