@@ -17,6 +17,8 @@ var commandConfigPairs = []configCommandPair{
 	{Config: AcknowledgementsConfig{}, Command: AcknowledgementsCommand},
 	{Config: AgentStartConfig{}, Command: AgentStartCommand},
 	{Config: AgentStopConfig{}, Command: AgentStopCommand},
+	{Config: AgentPauseConfig{}, Command: AgentPauseCommand},
+	{Config: AgentResumeConfig{}, Command: AgentResumeCommand},
 	{Config: AnnotateConfig{}, Command: AnnotateCommand},
 	{Config: AnnotationRemoveConfig{}, Command: AnnotationRemoveCommand},
 	{Config: ArtifactDownloadConfig{}, Command: ArtifactDownloadCommand},
@@ -30,6 +32,7 @@ var commandConfigPairs = []configCommandPair{
 	{Config: EnvSetConfig{}, Command: EnvSetCommand},
 	{Config: EnvUnsetConfig{}, Command: EnvUnsetCommand},
 	{Config: GitCredentialsHelperConfig{}, Command: GitCredentialsHelperCommand},
+	{Config: KubernetesBootstrapConfig{}, Command: KubernetesBootstrapCommand},
 	{Config: LockAcquireConfig{}, Command: LockAcquireCommand},
 	{Config: LockDoConfig{}, Command: LockDoCommand},
 	{Config: LockDoneConfig{}, Command: LockDoneCommand},
@@ -59,7 +62,7 @@ func TestAllCommandConfigStructsHaveCorrespondingCLIFlags(t *testing.T) {
 			flagNames[flag.GetName()] = struct{}{}
 		}
 
-		fields, err := reflections.Fields(pair.Config)
+		fields, err := reflections.FieldsDeep(pair.Config)
 		if err != nil {
 			t.Fatalf("getting fields for type %T: %v", pair.Config, err)
 		}
