@@ -92,6 +92,13 @@ var (
 		EnvVar: "BUILDKITE_STRICT_SINGLE_HOOKS",
 	}
 
+	SocketsPathFlag = cli.StringFlag{
+		Name:   "sockets-path",
+		Value:  defaultSocketsPath(),
+		Usage:  "Directory where the agent will place sockets",
+		EnvVar: "BUILDKITE_SOCKETS_PATH",
+	}
+
 	KubernetesExecFlag = cli.BoolFlag{
 		Name: "kubernetes-exec",
 		Usage: "This is intended to be used only by the Buildkite k8s stack " +
@@ -100,6 +107,14 @@ var (
 		EnvVar: "BUILDKITE_KUBERNETES_EXEC",
 	}
 
+	KubernetesContainerIDFlag = cli.IntFlag{
+		Name: "kubernetes-container-id",
+		Usage: "This is intended to be used only by the Buildkite k8s stack " +
+			"(github.com/buildkite/agent-stack-k8s); it sets an ID number " +
+			"used to identify this container within the pod",
+		EnvVar: "BUILDKITE_CONTAINER_ID",
+	}
+
 	NoMultipartArtifactUploadFlag = cli.BoolFlag{
 		Name:   "no-multipart-artifact-upload",
 		Usage:  "For Buildkite-hosted artifacts, disables the use of multipart uploads. Has no effect on uploads to other destinations such as custom cloud buckets",
@@ -138,6 +153,27 @@ var (
 	}
 )
 
+// GlobalConfig includes very common shared config options for easy inclusion across
+// config structs (via embedding).
+type GlobalConfig struct {
+	Debug       bool     `cli:"debug"`
+	LogLevel    string   `cli:"log-level"`
+	NoColor     bool     `cli:"no-color"`
+	Experiments []string `cli:"experiment" normalize:"list"`
+	Profile     string   `cli:"profile"`
+}
+
+// APIConfig includes API-related shared options for easy inclusion across
+// config structs (via embedding). Subcommands that don't need APIConfig usually
+// do something "trivial" (e.g. acknowledgements) or "special" (e.g. start).
+type APIConfig struct {
+	AgentAccessToken string `cli:"agent-access-token" validate:"required"`
+	DebugHTTP        bool   `cli:"debug-http"`
+	TraceHTTP        bool   `cli:"trace-http"`
+	Endpoint         string `cli:"endpoint" validate:"required"`
+	NoHTTP2          bool   `cli:"no-http2"`
+}
+
 func globalFlags() []cli.Flag {
 	return []cli.Flag{
 		NoColorFlag,
@@ -148,6 +184,16 @@ func globalFlags() []cli.Flag {
 	}
 }
 
+func apiFlags() []cli.Flag {
+	return []cli.Flag{
+		AgentAccessTokenFlag,
+		EndpointFlag,
+		NoHTTP2Flag,
+		DebugHTTPFlag,
+		TraceHTTPFlag,
+	}
+}
+
 func CreateLogger(cfg any) logger.Logger {
 	var l logger.Logger
 	logFormat := "text"
