@@ -1,20 +1,22 @@
 package clicommand
 
 import (
+	"cmp"
 	"context"
 	"encoding/json"
 	"errors"
 	"fmt"
 	"io"
+	"iter"
 	"maps"
 	"os"
 	"os/exec"
-	"path"
 	"path/filepath"
 	"slices"
 	"strings"
 	"time"
 
+	"drjosh.dev/zzglob"
 	"github.com/aws/aws-sdk-go-v2/service/kms"
 	"github.com/buildkite/agent/v3/agent"
 	"github.com/buildkite/agent/v3/api"
@@ -68,8 +70,15 @@ Example:
     $ buildkite-agent pipeline upload my-custom-pipeline.yml
     $ ./script/dynamic_step_generator | buildkite-agent pipeline upload`
 
+const ifChangedSkippedMsg = "if_changed pattern did not match any paths changed in this build"
+
+const defaultGitDiffBase = "origin/main"
+
 type PipelineUploadConfig struct {
-	FilePath        string   `cli:"arg:0" label:"upload paths"`
+	GlobalConfig
+	APIConfig
+
+	FilePaths       []string `cli:"arg:*" label:"upload paths"`
 	Replace         bool     `cli:"replace"`
 	Job             string   `cli:"job"` // required, but not in dry-run mode
 	DryRun          bool     `cli:"dry-run"`
@@ -78,31 +87,22 @@ type PipelineUploadConfig struct {
 	RedactedVars    []string `cli:"redacted-vars" normalize:"list"`
 	RejectSecrets   bool     `cli:"reject-secrets"`
 
+	// Used for if_changed processing
+	ApplyIfChanged bool   `cli:"apply-if-changed"`
+	GitDiffBase    string `cli:"git-diff-base"`
+
 	// Used for signing
 	JWKSFile         string `cli:"jwks-file"`
 	JWKSKeyID        string `cli:"jwks-key-id"`
 	SigningAWSKMSKey string `cli:"signing-aws-kms-key"`
 	DebugSigning     bool   `cli:"debug-signing"`
-
-	// Global flags
-	Debug       bool     `cli:"debug"`
-	LogLevel    string   `cli:"log-level"`
-	NoColor     bool     `cli:"no-color"`
-	Experiments []string `cli:"experiment" normalize:"list"`
-	Profile     string   `cli:"profile"`
-
-	// API config
-	DebugHTTP        bool   `cli:"debug-http"`
-	AgentAccessToken string `cli:"agent-access-token"` // required, but not in dry-run mode
-	Endpoint         string `cli:"endpoint" validate:"required"`
-	NoHTTP2          bool   `cli:"no-http2"`
 }
 
 var PipelineUploadCommand = cli.Command{
 	Name:        "upload",
 	Usage:       "Uploads a description of a build pipeline adds it to the currently running build after the current job",
 	Description: pipelineUploadHelpDescription,
-	Flags: []cli.Flag{
+	Flags: slices.Concat(globalFlags(), apiFlags(), []cli.Flag{
 		cli.BoolFlag{
 			Name:   "replace",
 			Usage:  "Replace the rest of the existing pipeline with the steps uploaded. Jobs that are already running are not removed.",
@@ -135,6 +135,16 @@ var PipelineUploadCommand = cli.Command{
 			Usage:  "When true, fail the pipeline upload early if the pipeline contains secrets",
 			EnvVar: "BUILDKITE_AGENT_PIPELINE_UPLOAD_REJECT_SECRETS",
 		},
+		cli.BoolTFlag{
+			Name:   "apply-if-changed",
+			Usage:  "When enabled, steps containing an ′if_changed′ key are evaluated against the git diff. If the ′if_changed′ glob pattern match no files changed in the build, the step is skipped.",
+			EnvVar: "BUILDKITE_AGENT_APPLY_SKIP_IF_UNCHANGED",
+		},
+		cli.StringFlag{
+			Name:   "git-diff-base",
+			Usage:  "Provides the base from which to find the git diff when processing ′if_changed′, e.g. origin/main. If not provided, it uses the first valid value of {origin/$BUILDKITE_PULL_REQUEST_BASE_BRANCH, origin/$BUILDKITE_PIPELINE_DEFAULT_BRANCH, origin/main}.",
+			EnvVar: "BUILDKITE_GIT_DIFF_BASE",
+		},
 
 		// Note: changes to these environment variables need to be reflected in the environment created
 		// in the job runner. At the momenet, that's at agent/job_runner.go:500-507
@@ -158,47 +168,38 @@ var PipelineUploadCommand = cli.Command{
 			Usage:  "Enable debug logging for pipeline signing. This can potentially leak secrets to the logs as it prints each step in full before signing. Requires debug logging to be enabled",
 			EnvVar: "BUILDKITE_AGENT_DEBUG_SIGNING",
 		},
-
-		// API Flags
-		AgentAccessTokenFlag,
-		EndpointFlag,
-		NoHTTP2Flag,
-		DebugHTTPFlag,
-
-		// Global flags
-		NoColorFlag,
-		DebugFlag,
-		LogLevelFlag,
-		ExperimentsFlag,
-		ProfileFlag,
 		RedactedVars,
-	},
+	}),
 	Action: func(c *cli.Context) error {
 		ctx := context.Background()
 		ctx, cfg, l, _, done := setupLoggerAndConfig[PipelineUploadConfig](ctx, c)
 		defer done()
 
-		// Find the pipeline either from STDIN or the first argument
-		var input *os.File
-		var filename string
+		// Find the pipeline either from STDIN or the non-flag arguments
+		type input struct {
+			file *os.File
+			name string
+		}
+		var inputs []input
 
 		switch {
-		case cfg.FilePath != "":
-			l.Info("Reading pipeline config from %q", cfg.FilePath)
+		case len(cfg.FilePaths) > 0:
+			l.Info("Reading pipeline configs from %q", cfg.FilePaths)
 
-			filename = filepath.Base(cfg.FilePath)
-			file, err := os.Open(cfg.FilePath)
-			if err != nil {
-				return fmt.Errorf("failed to read file: %w", err)
+			for _, fn := range cfg.FilePaths {
+				file, err := os.Open(fn)
+				if err != nil {
+					return fmt.Errorf("failed to read file: %w", err)
+				}
+				defer file.Close()
+				inputs = append(inputs, input{file, filepath.Base(fn)})
 			}
-			defer file.Close()
-			input = file
 
 		case stdin.IsReadable():
 			l.Info("Reading pipeline config from STDIN")
 
 			// Actually read the file from STDIN
-			input = os.Stdin
+			inputs = []input{{os.Stdin, "(stdin)"}}
 
 		default:
 			l.Info("Searching for pipeline config...")
@@ -237,23 +238,25 @@ var PipelineUploadCommand = cli.Command{
 			l.Info("Found config file %q", found)
 
 			// Read the default file
-			filename = path.Base(found)
 			file, err := os.Open(found)
 			if err != nil {
 				return fmt.Errorf("failed to read file %q: %w", found, err)
 			}
 			defer file.Close()
-			input = file
+			inputs = []input{{file, filepath.Base(found)}}
 		}
 
-		// Make sure the file actually has something in it
-		if input != os.Stdin {
-			fi, err := input.Stat()
+		// Make sure each file (other than stdin) actually has something in it
+		for _, input := range inputs {
+			if input.file == os.Stdin {
+				continue
+			}
+			fi, err := input.file.Stat()
 			if err != nil {
-				return fmt.Errorf("couldn't stat pipeline configuration file %q: %w", input.Name(), err)
+				return fmt.Errorf("couldn't stat pipeline configuration file %q: %w", input.file.Name(), err)
 			}
 			if fi.Size() == 0 {
-				return fmt.Errorf("pipeline file %q is empty", input.Name())
+				return fmt.Errorf("pipeline file %q is empty", input.file.Name())
 			}
 		}
 
@@ -264,118 +267,144 @@ var PipelineUploadCommand = cli.Command{
 			resolveCommit(l, environ)
 		}
 
-		src := filename
-		if src == "" {
-			src = "(stdin)"
-		}
-
-		result, err := cfg.parseAndInterpolate(ctx, src, input, environ)
-		if err != nil {
-			w := warning.As(err)
-			if w == nil {
-				return err
-			}
-			l.Warn("There were some issues with the pipeline input - pipeline upload will proceed, but might not succeed:\n%v", w)
-		}
-
-		if len(cfg.RedactedVars) > 0 {
-			// Secret detection uses the original environment, since
-			// Interpolate merges the pipeline's env block into `environ`.
-			searchForSecrets(l, &cfg, environ, result, src)
-		}
-
-		var (
-			key signature.Key
-		)
-
-		switch {
-		case cfg.SigningAWSKMSKey != "":
-			awscfg, err := awslib.GetConfigV2(ctx)
-			if err != nil {
-				return err
-			}
-
-			// assign a crypto signer which uses the KMS key to sign the pipeline
-			key, err = awssigner.NewKMS(kms.NewFromConfig(awscfg), cfg.SigningAWSKMSKey)
-			if err != nil {
-				return fmt.Errorf("couldn't create KMS signer: %w", err)
-			}
-
-		case cfg.JWKSFile != "":
-			key, err = jwkutil.LoadKey(cfg.JWKSFile, cfg.JWKSKeyID)
-			if err != nil {
-				return fmt.Errorf("couldn't read the signing key file: %w", err)
-			}
-		}
-
-		if key != nil {
-
-			err = signature.SignSteps(
-				ctx,
-				result.Steps,
-				key,
-				os.Getenv("BUILDKITE_REPO"),
-				signature.WithEnv(result.Env.ToMap()),
-				signature.WithLogger(l),
-				signature.WithDebugSigning(cfg.DebugSigning),
-			)
-			if err != nil {
-				return fmt.Errorf("couldn't sign pipeline: %w", err)
-			}
-		}
-
-		// In dry-run mode we just output the generated pipeline to stdout.
+		// Used to encode output in dry-run mode.
+		dryRunEnc := func(any) error { return nil }
 		if cfg.DryRun {
-			var encode func(any) error
-
 			switch cfg.DryRunFormat {
 			case "json":
 				enc := json.NewEncoder(c.App.Writer)
 				enc.SetIndent("", "  ")
-				encode = enc.Encode
+				dryRunEnc = enc.Encode
 
 			case "yaml":
-				encode = yaml.NewEncoder(c.App.Writer).Encode
+				dryRunEnc = yaml.NewEncoder(c.App.Writer).Encode
 
 			default:
 				return fmt.Errorf("unknown output format %q", cfg.DryRunFormat)
 			}
+		}
 
-			// All logging happens to stderr.
-			// So this can be used with other tools to get interpolated, signed
-			// JSON or YAML.
-			if err := encode(result); err != nil {
-				return err
+		prependOriginIfNonempty := func(key string) string {
+			s := os.Getenv(key)
+			if s == "" {
+				return ""
 			}
-
-			return nil
+			return "origin/" + s
 		}
 
-		// Check we have a job id set if not in dry run
-		if cfg.Job == "" {
-			return errors.New("missing job parameter. Usually this is set in the environment for a Buildkite job via BUILDKITE_JOB_ID.")
+		ifChanged := &ifChangedApplicator{
+			enabled: cfg.ApplyIfChanged,
+			diffBase: cmp.Or(
+				cfg.GitDiffBase,
+				prependOriginIfNonempty("BUILDKITE_PULL_REQUEST_BASE_BRANCH"),
+				prependOriginIfNonempty("BUILDKITE_PIPELINE_DEFAULT_BRANCH"),
+				defaultGitDiffBase,
+			),
 		}
 
-		// Check we have an agent access token if not in dry run
-		if cfg.AgentAccessToken == "" {
-			return errors.New("missing agent-access-token parameter. Usually this is set in the environment for a Buildkite job via BUILDKITE_AGENT_ACCESS_TOKEN.")
-		}
+		// Process all inputs.
+		for _, input := range inputs {
 
-		uploader := &agent.PipelineUploader{
-			Client: api.NewClient(l, loadAPIClientConfig(cfg, "AgentAccessToken")),
-			JobID:  cfg.Job,
-			Change: &api.PipelineChange{
-				UUID:     api.NewUUID(),
-				Replace:  cfg.Replace,
-				Pipeline: result,
-			},
-			RetrySleepFunc: time.Sleep,
-		}
-		if err := uploader.Upload(ctx, l); err != nil {
-			return err
-		}
+			// For each pipeline in the input (could be multiple)...
+			count := 1
+			for result, err := range cfg.parseAndInterpolate(ctx, input.name, input.file, environ) {
+				if err != nil {
+					w := warning.As(err)
+					if w == nil {
+						return err
+					}
+					l.Warn("There were some issues with the pipeline input - pipeline upload will proceed, but might not succeed:\n%v", w)
+				}
 
-		l.Info("Successfully uploaded and parsed pipeline config")
+				if len(cfg.RedactedVars) > 0 {
+					// Secret detection uses the original environment, since
+					// Interpolate merges the pipeline's env block into `environ`.
+					searchForSecrets(l, &cfg, environ, result, input.name)
+				}
+
+				var (
+					key signature.Key
+				)
+
+				switch {
+				case cfg.SigningAWSKMSKey != "":
+					awscfg, err := awslib.GetConfigV2(ctx)
+					if err != nil {
+						return err
+					}
+
+					// assign a crypto signer which uses the KMS key to sign the pipeline
+					key, err = awssigner.NewKMS(kms.NewFromConfig(awscfg), cfg.SigningAWSKMSKey)
+					if err != nil {
+						return fmt.Errorf("couldn't create KMS signer: %w", err)
+					}
+
+				case cfg.JWKSFile != "":
+					key, err = jwkutil.LoadKey(cfg.JWKSFile, cfg.JWKSKeyID)
+					if err != nil {
+						return fmt.Errorf("couldn't read the signing key file: %w", err)
+					}
+				}
+
+				if key != nil {
+					err := signature.SignSteps(
+						ctx,
+						result.Steps,
+						key,
+						os.Getenv("BUILDKITE_REPO"),
+						signature.WithEnv(result.Env.ToMap()),
+						signature.WithLogger(l),
+						signature.WithDebugSigning(cfg.DebugSigning),
+					)
+					if err != nil {
+						return fmt.Errorf("couldn't sign pipeline: %w", err)
+					}
+				}
+
+				// Apply or strip out `if_changed`, based on settings.
+				ifChanged.apply(l, result.Steps)
+
+				// All logging happens to stderr.
+				// So this can be used with other tools to get interpolated, signed
+				// JSON or YAML.
+				if err := dryRunEnc(result); err != nil {
+					return err
+				}
+
+				// In dry-run mode we just output the generated pipeline to stdout,
+				// and don't want to upload it.
+				if cfg.DryRun {
+					continue
+				}
+
+				// Check we have a job id set if not in dry run
+				if cfg.Job == "" {
+					return errors.New("missing job parameter. Usually this is set in the environment for a Buildkite job via BUILDKITE_JOB_ID.")
+				}
+
+				// Check we have an agent access token if not in dry run
+				if cfg.AgentAccessToken == "" {
+					return errors.New("missing agent-access-token parameter. Usually this is set in the environment for a Buildkite job via BUILDKITE_AGENT_ACCESS_TOKEN.")
+				}
+
+				uploader := &agent.PipelineUploader{
+					Client: api.NewClient(l, loadAPIClientConfig(cfg, "AgentAccessToken")),
+					JobID:  cfg.Job,
+					Change: &api.PipelineChange{
+						UUID:     api.NewUUID(),
+						Replace:  cfg.Replace,
+						Pipeline: result,
+					},
+					RetrySleepFunc: time.Sleep,
+				}
+				if err := uploader.Upload(ctx, l); err != nil {
+					return err
+				}
+
+				l.Info("Successfully parsed and uploaded pipeline #%d from %q", count, input.name)
+				count++
+			}
+		}
 
 		return nil
 	},
@@ -543,26 +572,184 @@ func searchForSecrets(
 	return nil
 }
 
-func (cfg *PipelineUploadConfig) parseAndInterpolate(ctx context.Context, src string, input io.Reader, environ *env.Environment) (*pipeline.Pipeline, error) {
-	result, err := pipeline.Parse(input)
-	if err != nil && !warning.Is(err) {
-		return nil, fmt.Errorf("pipeline parsing of %q failed: %w", src, err)
-	}
-	if cfg.NoInterpolation {
-		// Note that err may be nil or a non-nil warning from pipeline.Parse
-		return result, err
-	}
-	// Pass the trace context from our environment to the pipeline.
-	if tracing, has := environ.Get(tracetools.EnvVarTraceContextKey); has {
-		if result.Env == nil {
-			result.Env = ordered.NewMap[string, string](1)
+func (cfg *PipelineUploadConfig) parseAndInterpolate(ctx context.Context, src string, input io.Reader, environ *env.Environment) iter.Seq2[*pipeline.Pipeline, error] {
+	return func(yield func(*pipeline.Pipeline, error) bool) {
+		for result, err := range pipeline.ParseAll(input) {
+			// Check the error, apply interpolation if needed.
+			switch {
+			case err != nil && !warning.Is(err):
+				err = fmt.Errorf("pipeline parsing of %q failed: %w", src, err)
+				// yield below
+
+			case cfg.NoInterpolation:
+				// Note that err may be nil or a non-nil warning from pipeline.Parse
+				// yield below
+
+			default: // yes, interpolation
+				// Pass the trace context from our environment to the pipeline.
+				if tracing, has := environ.Get(tracetools.EnvVarTraceContextKey); has {
+					if result.Env == nil {
+						result.Env = ordered.NewMap[string, string](1)
+					}
+					result.Env.Set(tracetools.EnvVarTraceContextKey, tracing)
+				}
+
+				// Do the interpolation.
+				preferRuntimeEnv := experiments.IsEnabled(ctx, experiments.InterpolationPrefersRuntimeEnv)
+				err = result.Interpolate(environ, preferRuntimeEnv)
+				if err != nil {
+					err = fmt.Errorf("pipeline interpolation of %q failed: %w", src, err)
+				}
+				// yield below
+			}
+
+			// Yield the result and error.
+			if !yield(result, err) {
+				return
+			}
 		}
-		result.Env.Set(tracetools.EnvVarTraceContextKey, tracing)
 	}
-	// Do the interpolation.
-	preferRuntimeEnv := experiments.IsEnabled(ctx, experiments.InterpolationPrefersRuntimeEnv)
-	if err := result.Interpolate(environ, preferRuntimeEnv); err != nil {
-		return nil, fmt.Errorf("pipeline interpolation of %q failed: %w", src, err)
+}
+
+// gatherChangedFiles determines changed files in this build.
+func gatherChangedFiles(diffBase string) (mergeBase string, changedPaths []string, err error) {
+	// The --merge-base flag was only added to git-diff recently.
+	mergeBaseOut, err := exec.Command("git", "merge-base", diffBase, "HEAD").Output()
+	if err != nil {
+		return "", nil, gitMergeBaseError{diffBase: diffBase, wrapped: err}
 	}
-	return result, err
+	mergeBase = strings.TrimSpace(string(mergeBaseOut))
+
+	gitDiff, err := exec.Command("git", "diff", "--name-only", mergeBase).Output()
+	if err != nil {
+		return mergeBase, nil, gitDiffError{mergeBase: mergeBase, wrapped: err}
+	}
+	changedPaths = strings.Split(string(gitDiff), "\n")
+	changedPaths = slices.DeleteFunc(changedPaths, func(s string) bool {
+		return strings.TrimSpace(s) == ""
+	})
+	return mergeBase, changedPaths, nil
+}
+
+type gitMergeBaseError struct {
+	diffBase string
+	wrapped  error
+}
+
+func (e gitMergeBaseError) Error() string {
+	return fmt.Sprintf("git merge-base %q HEAD: %v", e.diffBase, e.wrapped)
+}
+func (e gitMergeBaseError) Unwrap() error { return e.wrapped }
+
+type gitDiffError struct {
+	mergeBase string
+	wrapped   error
+}
+
+func (e gitDiffError) Error() string {
+	return fmt.Sprintf("git diff --name-only %q: %v", e.mergeBase, e.wrapped)
+}
+func (e gitDiffError) Unwrap() error { return e.wrapped }
+
+// ifChangedApplicator applies `if_changed` as it appears within the pipeline
+// being uploaded. The `if_changed` attribute takes a glob pattern of files
+// to match. The step is skipped if the glob doesn't match any "changed files".
+type ifChangedApplicator struct {
+	enabled      bool // apply-if-changed is enabled
+	gathered     bool // the changed files have been computed?
+	diffBase     string
+	changedPaths []string
+}
+
+// apply applies "if_changed". If it's not enabled, it strips "if_changed"
+// attributes. Otherwise, it converts them into "skip" if the glob
+// pattern matches no changed files.
+func (ica *ifChangedApplicator) apply(l logger.Logger, steps pipeline.Steps) {
+stepsLoop:
+	for _, step := range steps {
+		// All supported step types store "if_changed" in a map.
+		var content map[string]any
+
+		// These step types support "skip".
+		switch st := step.(type) {
+		case *pipeline.GroupStep:
+			// Recurse into group steps
+			ica.apply(l, st.Steps)
+			content = st.RemainingFields
+
+		case *pipeline.CommandStep:
+			content = st.RemainingFields
+
+		case *pipeline.TriggerStep:
+			content = st.Contents
+		}
+
+		// Retrieve the value then delete it.
+		// It's not yet understood by the backend.
+		ic := content["if_changed"]
+		if ic == nil {
+			continue
+		}
+		delete(content, "if_changed")
+
+		if !ica.enabled {
+			// Not applying the if_changed; leaving it deleted.
+			continue
+		}
+
+		// If we don't know the changed paths yet, call out to Git.
+		if !ica.gathered {
+			mergeBase, cps, err := gatherChangedFiles(ica.diffBase)
+			if err != nil {
+				l.Error("Couldn't determine git diff from upstream, not skipping any pipeline steps: %v", err)
+				switch err := err.(type) {
+				case gitMergeBaseError:
+					l.Error("This could be because %q might not be a commit in the repository.\n"+
+						"You may need to change the --git-diff-base flag or BUILDKITE_GIT_DIFF_BASE env var.",
+						err.diffBase,
+					)
+
+				case gitDiffError:
+					l.Error("This could be because the merge-base that Git found, %q, might be invalid.\n"+
+						"You may need to change the --git-diff-base flag or BUILDKITE_GIT_DIFF_BASE env var.",
+						err.mergeBase,
+					)
+				}
+
+				// Because changed files couldn't be determined, we switch into
+				// disabled mode.
+				ica.enabled = false
+				continue stepsLoop
+			}
+
+			// The changed files are now known.
+			l.Info("Applying if_changed conditions relative to %q (the merge-base of %q and HEAD) - %d changed files", mergeBase, ica.diffBase, len(cps))
+			ica.gathered = true
+			ica.changedPaths = cps
+		}
+
+		// Parse and test the glob pattern against the paths.
+		pattern, ok := ic.(string)
+		if !ok {
+			l.Warn("if_changed value must be a string containing a glob pattern (was a %T)\n"+
+				"The step will not be skipped.", ic)
+			continue
+		}
+
+		glob, err := zzglob.Parse(pattern)
+		if err != nil {
+			l.Warn("if_changed value %q couldn't be parsed as a glob pattern: %v\n"+
+				"The step will not be skipped.", pattern, err)
+			continue
+		}
+		for _, cp := range ica.changedPaths {
+			if glob.Match(cp) {
+				continue stepsLoop // one of the paths changed
+			}
+		}
+
+		// Glob matched no changed file paths - this step is now skipped.
+		// Note that the "skip" string is limited to 70 characters.
+		content["skip"] = ifChangedSkippedMsg
+	}
 }
