@@ -24,7 +24,6 @@ import (
 	"github.com/buildkite/agent/v3/logger"
 	"github.com/buildkite/roko"
 	"github.com/dustin/go-humanize"
-	"github.com/mattn/go-zglob"
 )
 
 const (
@@ -159,6 +158,7 @@ func (a *Uploader) collect(ctx context.Context) ([]*api.Artifact, error) {
 	// workers will avoid slamming the runtime (as could happen with a
 	// goroutine per file).
 	wctx, cancel := context.WithCancelCause(ctx)
+	defer cancel(nil)
 	var wg sync.WaitGroup
 	for range runtime.GOMAXPROCS(0) {
 		wg.Add(1)
@@ -200,65 +200,37 @@ func (a *Uploader) glob(ctx context.Context, filesCh chan<- string) error {
 	// glob is solely responsible for writing to the channel.
 	defer close(filesCh)
 
-	if experiments.IsEnabled(ctx, experiments.UseZZGlob) {
-		// New zzglob library. Do all globs at once with MultiGlob, which takes
-		// care of any necessary parallelism under the hood.
-		a.logger.Debug("Searching for %s", a.conf.Paths)
-		var patterns []*zzglob.Pattern
-		for _, globPath := range strings.Split(a.conf.Paths, ArtifactPathDelimiter) {
-			globPath := strings.TrimSpace(globPath)
-			if globPath == "" {
-				continue
-			}
-			pattern, err := zzglob.Parse(globPath)
-			if err != nil {
-				return fmt.Errorf("invalid glob pattern %q: %w", globPath, err)
-			}
-			patterns = append(patterns, pattern)
-		}
-
-		walkDirFunc := func(path string, d fs.DirEntry, err error) error {
-			if err != nil {
-				a.logger.Warn("Couldn't walk path %s", path)
-				return nil
-			}
-			if d != nil && d.IsDir() {
-				a.logger.Warn("One of the glob patterns matched a directory: %s", path)
-				return nil
-			}
-			filesCh <- path
-			return nil
-		}
-		err := zzglob.MultiGlob(ctx, patterns, walkDirFunc, zzglob.TraverseSymlinks(a.conf.GlobResolveFollowSymlinks))
-		if err != nil {
-			return fmt.Errorf("globbing patterns: %w", err)
-		}
-		return nil
-	}
-
-	// Old go-zglob library. Do each glob one at a time.
-	// go-zglob uses fastwalk under the hood to parallelise directory walking.
-	globfunc := zglob.Glob
-	if a.conf.GlobResolveFollowSymlinks {
-		// Follow symbolic links for files & directories while expanding globs
-		globfunc = zglob.GlobFollowSymlinks
-	}
+	// New zzglob library. Do all globs at once with MultiGlob, which takes
+	// care of any necessary parallelism under the hood.
+	a.logger.Debug("Searching for %s", a.conf.Paths)
+	var patterns []*zzglob.Pattern
 	for _, globPath := range strings.Split(a.conf.Paths, ArtifactPathDelimiter) {
 		globPath := strings.TrimSpace(globPath)
 		if globPath == "" {
 			continue
 		}
-		files, err := globfunc(globPath)
-		if errors.Is(err, os.ErrNotExist) {
-			a.logger.Info("File not found: %s", globPath)
-			continue
-		}
+		pattern, err := zzglob.Parse(globPath)
 		if err != nil {
-			return fmt.Errorf("resolving glob %s: %w", globPath, err)
+			return fmt.Errorf("invalid glob pattern %q: %w", globPath, err)
 		}
-		for _, path := range files {
-			filesCh <- path
+		patterns = append(patterns, pattern)
+	}
+
+	walkDirFunc := func(path string, d fs.DirEntry, err error) error {
+		if err != nil {
+			a.logger.Warn("Couldn't walk path %s", path)
+			return nil
 		}
+		if d != nil && d.IsDir() {
+			a.logger.Warn("One of the glob patterns matched a directory: %s", path)
+			return nil
+		}
+		filesCh <- path
+		return nil
+	}
+	err := zzglob.MultiGlob(ctx, patterns, walkDirFunc, zzglob.TraverseSymlinks(a.conf.GlobResolveFollowSymlinks))
+	if err != nil {
+		return fmt.Errorf("globbing patterns: %w", err)
 	}
 	return nil
 }
@@ -350,7 +322,7 @@ func (a *Uploader) build(path string, absolutePath string) (*api.Artifact, error
 	if err != nil {
 		return nil, fmt.Errorf("opening file %s: %w", absolutePath, err)
 	}
-	defer file.Close()
+	defer file.Close() //nolint:errcheck // File is only open for read.
 
 	// Generate a SHA-1 and SHA-256 checksums for the file.
 	// Writing to hashes never errors, but reading from the file might.
@@ -652,12 +624,6 @@ func (a *artifactUploadWorker) doWorkUnits(ctx context.Context, unitsCh <-chan w
 
 func (a *artifactUploadWorker) stateUpdater(ctx context.Context, resultsCh <-chan workUnitResult, stateUpdaterErrCh chan<- error) {
 	var errs []error
-	defer func() {
-		if len(errs) > 0 {
-			stateUpdaterErrCh <- errors.Join(errs...)
-		}
-		close(stateUpdaterErrCh)
-	}()
 
 	// When this ticks, upload any pending artifact states as a batch.
 	updateTicker := time.NewTicker(1 * time.Second)
@@ -666,7 +632,7 @@ selectLoop:
 	for {
 		select {
 		case <-ctx.Done():
-			return
+			break selectLoop
 
 		case <-updateTicker.C:
 			// Note: updateStates removes trackers for completed states.
@@ -711,7 +677,11 @@ selectLoop:
 	if err := a.updateStates(ctx); err != nil {
 		errs = append(errs, err)
 	}
-	return
+
+	if len(errs) > 0 {
+		stateUpdaterErrCh <- errors.Join(errs...)
+	}
+	close(stateUpdaterErrCh)
 }
 
 // updateStates uploads terminal artifact states to Buildkite in a batch.
