@@ -122,13 +122,13 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 	// Check if not nil to allow for tests to overwrite shell.
 	if e.shell == nil {
 		sh, err := shell.New(
-			shell.WithDebug(e.ExecutorConfig.Debug),
+			shell.WithDebug(e.Debug),
 			shell.WithEnv(environ),
 			shell.WithLogger(preRedactedLogger), // shell -> logger -> redactor -> real stderr
-			shell.WithInterruptSignal(e.ExecutorConfig.CancelSignal),
-			shell.WithPTY(e.ExecutorConfig.RunInPty),
+			shell.WithInterruptSignal(e.CancelSignal),
+			shell.WithPTY(e.RunInPty),
 			shell.WithStdout(preRedactedStdout), // shell -> redactor -> real stdout
-			shell.WithSignalGracePeriod(e.ExecutorConfig.SignalGracePeriod),
+			shell.WithSignalGracePeriod(e.SignalGracePeriod),
 			shell.WithTraceContextCodec(e.TraceContextCodec),
 		)
 		if err != nil {
@@ -145,7 +145,7 @@ func (e *Executor) Run(ctx context.Context) (exitCode int) {
 
 	// Listen for cancellation. Once ctx is cancelled, some tasks can run
 	// afterwards during the signal grace period. These use graceCtx.
-	graceCtx, graceCancel := withGracePeriod(ctx, e.SignalGracePeriod)
+	graceCtx, graceCancel := WithGracePeriod(ctx, e.SignalGracePeriod)
 	defer graceCancel()
 	go func() {
 		<-e.cancelCh
@@ -309,10 +309,17 @@ func (e *Executor) Cancel() error {
 		return errors.New("already cancelled")
 	}
 	e.cancelled = true
+	e.shell.Env.Set("BUILDKITE_JOB_CANCELLED", "true")
 	close(e.cancelCh)
 	return nil
 }
 
+const (
+	HookScopeAgent      = "agent"
+	HookScopeRepository = "repository"
+	HookScopePlugin     = "plugin"
+)
+
 type HookConfig struct {
 	Name           string
 	Scope          string
@@ -323,18 +330,17 @@ type HookConfig struct {
 }
 
 func (e *Executor) tracingImplementationSpecificHookScope(scope string) string {
-	if e.TracingBackend != tracetools.BackendOpenTelemetry {
+	if e.TracingBackend != tracetools.BackendDatadog {
 		return scope
 	}
 
-	// The scope names local and global are confusing, and different to what we document, so we should use the
-	// documented names (repository and agent, respectively) in OpenTelemetry.
-	// However, we need to keep the OpenTracing/Datadog implementation the same, hence this horrible function
+	// In olden times, when the datadog tracing backend was written, these hook scopes were named "local" and "global"
+	// We need to maintain backwards compatibility with the old names for span attribute reasons, so we map them here
 	switch scope {
-	case "local":
-		return "repository"
-	case "global":
-		return "agent"
+	case HookScopeRepository:
+		return "local"
+	case HookScopeAgent:
+		return "global"
 	default:
 		return scope
 	}
@@ -344,7 +350,7 @@ func (e *Executor) tracingImplementationSpecificHookScope(scope string) string {
 func (e *Executor) executeHook(ctx context.Context, hookCfg HookConfig) error {
 	scopeName := e.tracingImplementationSpecificHookScope(hookCfg.Scope)
 	spanName := e.implementationSpecificSpanName(fmt.Sprintf("%s %s hook", scopeName, hookCfg.Name), "hook.execute")
-	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.TracingBackend)
 	var err error
 	defer func() { span.FinishWithError(err) }()
 	span.AddAttributes(map[string]string{
@@ -419,7 +425,7 @@ Hooks of this kind are unfortunately not supported on Windows, as we have no way
 	}
 }
 
-func (e *Executor) runUnwrappedHook(ctx context.Context, hookName string, hookCfg HookConfig) error {
+func (e *Executor) runUnwrappedHook(ctx context.Context, _ string, hookCfg HookConfig) error {
 	environ := hookCfg.Env.Copy()
 
 	environ.Set("BUILDKITE_HOOK_PHASE", hookCfg.Name)
@@ -604,7 +610,7 @@ func (e *Executor) applyEnvironmentChanges(changes hook.EnvChanges) {
 	e.shell.Env.Apply(changes.Diff)
 
 	// reset output redactors based on new environment variable values
-	toRedact, short, err := redact.Vars(e.ExecutorConfig.RedactedVars, e.shell.Env.DumpPairs())
+	toRedact, short, err := redact.Vars(e.RedactedVars, e.shell.Env.DumpPairs())
 	if err != nil {
 		e.shell.OptionalWarningf("bad-redacted-vars", "Couldn't match environment variable names against redacted-vars: %v", err)
 	}
@@ -619,7 +625,7 @@ func (e *Executor) applyEnvironmentChanges(changes hook.EnvChanges) {
 
 	// First, let see any of the environment variables are supposed
 	// to change the job configuration at run time.
-	executorConfigEnvChanges := e.ExecutorConfig.ReadFromEnvironment(e.shell.Env)
+	executorConfigEnvChanges := e.ReadFromEnvironment(e.shell.Env)
 
 	// Print out the env vars that changed. As we go through each
 	// one, we'll determine if it was a special environment variable
@@ -698,7 +704,7 @@ func (e *Executor) executeGlobalHook(ctx context.Context, name string) error {
 	}
 	for _, h := range allHooks {
 		err = e.executeHook(ctx, HookConfig{
-			Scope: "global",
+			Scope: HookScopeAgent,
 			Name:  name,
 			Path:  h,
 		})
@@ -726,8 +732,8 @@ func (e *Executor) executeLocalHook(ctx context.Context, name string) error {
 	if err != nil {
 		if errors.Is(err, os.ErrNotExist) {
 			// If the hook doesn't exist, that's fine, we'll just skip it
-			if e.ExecutorConfig.Debug {
-				e.shell.Logger.Commentf("Local hook %s doesn't exist: %s, skipping", name, err)
+			if e.Debug {
+				e.shell.Commentf("Local hook %s doesn't exist: %s, skipping", name, err)
 			}
 			return nil
 		}
@@ -739,7 +745,7 @@ func (e *Executor) executeLocalHook(ctx context.Context, name string) error {
 	}
 
 	// For high-security configs, we allow the disabling of local hooks.
-	localHooksEnabled := e.ExecutorConfig.LocalHooksEnabled
+	localHooksEnabled := e.LocalHooksEnabled
 
 	// Allow hooks to disable local hooks by setting BUILDKITE_NO_LOCAL_HOOKS=true
 	noLocalHooks, _ := e.shell.Env.Get("BUILDKITE_NO_LOCAL_HOOKS")
@@ -752,7 +758,7 @@ func (e *Executor) executeLocalHook(ctx context.Context, name string) error {
 	}
 
 	return e.executeHook(ctx, HookConfig{
-		Scope: "local",
+		Scope: HookScopeRepository,
 		Name:  name,
 		Path:  localHookPath,
 	})
@@ -789,7 +795,7 @@ func addRepositoryHostToSSHKnownHosts(ctx context.Context, sh *shell.Shell, repo
 // setUp is run before all the phases run. It's responsible for initializing the
 // job environment
 func (e *Executor) setUp(ctx context.Context) error {
-	span, ctx := tracetools.StartSpanFromContext(ctx, "environment", e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, "environment", e.TracingBackend)
 	var err error
 	defer func() { span.FinishWithError(err) }()
 
@@ -797,7 +803,7 @@ func (e *Executor) setUp(ctx context.Context) error {
 	if e.BinPath != "" {
 		path, _ := e.shell.Env.Get("PATH")
 		// BinPath goes last so we don't disturb other tools
-		e.shell.Env.Set("PATH", fmt.Sprintf("%s%s%s", path, string(os.PathListSeparator), e.BinPath))
+		e.shell.Env.Set("PATH", fmt.Sprintf("%s%c%s", path, os.PathListSeparator, e.BinPath))
 	}
 
 	// Set a BUILDKITE_BUILD_CHECKOUT_PATH unless one exists already. We do this here
@@ -831,7 +837,7 @@ func (e *Executor) setUp(ctx context.Context) error {
 			if strings.HasPrefix(envar, "BUILDKITE_AGENT_ACCESS_TOKEN=") {
 				e.shell.Printf("BUILDKITE_AGENT_ACCESS_TOKEN=******************")
 			} else if strings.HasPrefix(envar, "BUILDKITE") || strings.HasPrefix(envar, "CI") || strings.HasPrefix(envar, "PATH") {
-				e.shell.Printf("%s", strings.Replace(envar, "\n", "\\n", -1))
+				e.shell.Printf("%s", strings.ReplaceAll(envar, "\n", "\\n"))
 			}
 		}
 	}
@@ -848,7 +854,7 @@ func (e *Executor) setUp(ctx context.Context) error {
 
 // tearDown is called before the executor exits, even on error
 func (e *Executor) tearDown(ctx context.Context) error {
-	span, ctx := tracetools.StartSpanFromContext(ctx, "pre-exit", e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, "pre-exit", e.TracingBackend)
 	var err error
 	defer func() { span.FinishWithError(err) }()
 
@@ -889,7 +895,7 @@ func (e *Executor) tearDown(ctx context.Context) error {
 // runPreCommandHooks runs the pre-command hooks and adds tracing spans.
 func (e *Executor) runPreCommandHooks(ctx context.Context) (err error) {
 	spanName := e.implementationSpecificSpanName("pre-command", "pre-command hooks")
-	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.TracingBackend)
 	defer func() { span.FinishWithError(err) }()
 
 	if err := e.executeGlobalHook(ctx, "pre-command"); err != nil {
@@ -919,7 +925,7 @@ func (e *Executor) runCommand(ctx context.Context) error {
 // runPostCommandHooks runs the post-command hooks and adds tracing spans.
 func (e *Executor) runPostCommandHooks(ctx context.Context) (err error) {
 	spanName := e.implementationSpecificSpanName("post-command", "post-command hooks")
-	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.TracingBackend)
 	defer func() { span.FinishWithError(err) }()
 
 	if err := e.executeGlobalHook(ctx, "post-command"); err != nil {
@@ -935,7 +941,7 @@ func (e *Executor) runPostCommandHooks(ctx context.Context) (err error) {
 func (e *Executor) CommandPhase(ctx context.Context) (hookErr error, commandErr error) {
 	var preCommandErr error
 
-	span, ctx := tracetools.StartSpanFromContext(ctx, "command", e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, "command", e.TracingBackend)
 	defer func() {
 		span.FinishWithError(hookErr)
 	}()
@@ -948,7 +954,7 @@ func (e *Executor) CommandPhase(ctx context.Context) (hookErr error, commandErr
 		}
 		// Because post-command hooks are often used for post-job cleanup, they
 		// can run during the grace period.
-		graceCtx, cancel := withGracePeriod(ctx, e.SignalGracePeriod)
+		graceCtx, cancel := WithGracePeriod(ctx, e.SignalGracePeriod)
 		defer cancel()
 		hookErr = e.runPostCommandHooks(graceCtx)
 	}()
@@ -981,17 +987,13 @@ func (e *Executor) CommandPhase(ctx context.Context) (hookErr error, commandErr
 
 	switch {
 	case isExitError && isExitSignaled:
-		// The recursive trap created a segfault that we were previously inadvertently suppressing
-		// in the next branch. Once the experiment is promoted, we should keep this branch in case
-		// to show the error to users.
 		e.shell.Errorf("The command was interrupted by a signal: %v", commandErr)
+		return nil, commandErr
 
-		// although error is an exit error, it's not returned. (seems like a bug)
-		// TODO: investigate phasing this out under a experiment
-		return nil, nil
 	case isExitError && !isExitSignaled:
 		e.shell.Errorf("The command exited with status %d", shell.ExitCode(commandErr))
 		return nil, commandErr
+
 	default:
 		e.shell.Errorf("%s", commandErr)
 
@@ -1005,7 +1007,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 	defer e.redactors.Flush()
 
 	spanName := e.implementationSpecificSpanName("default command hook", "hook.execute")
-	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.ExecutorConfig.TracingBackend)
+	span, ctx := tracetools.StartSpanFromContext(ctx, spanName, e.TracingBackend)
 	var err error
 	defer func() { span.FinishWithError(err) }()
 	span.AddAttributes(map[string]string{
@@ -1018,7 +1020,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 		return fmt.Errorf("The command phase has no `command` to execute. Provide a `command` field in your step configuration, or define a `command` hook in a step plug-in, your repository `.buildkite/hooks`, or agent `hooks-path`.")
 	}
 
-	scriptFileName := strings.Replace(e.Command, "\n", "", -1)
+	scriptFileName := strings.ReplaceAll(e.Command, "\n", "")
 	pathToCommand, err := filepath.Abs(filepath.Join(e.shell.Getwd(), scriptFileName))
 	commandIsScript := err == nil && osutil.FileExists(pathToCommand)
 	span.AddAttributes(map[string]string{"hook.command": pathToCommand})
@@ -1092,7 +1094,7 @@ func (e *Executor) defaultCommandPhase(ctx context.Context) error {
 		// command-eval agents, such risks are everpresent since the master
 		// can tell the agent to do anything anyway, but no-command-eval agents
 		// shouldn't be vulnerable to this!
-		if e.ExecutorConfig.CommandEval {
+		if e.CommandEval {
 			// Make script executable
 			if err = osutil.ChmodExecutable(pathToCommand); err != nil {
 				e.shell.Warningf("Error marking script %q as executable: %v", pathToCommand, err)
@@ -1207,7 +1209,7 @@ func (e *Executor) writeBatchScript(cmd string) (string, error) {
 //
 //	(returned shell.Logger) -> redactor 2 -> stderr
 func (e *Executor) setupRedactors(log shell.Logger, environ *env.Environment, stdout, stderr io.Writer) (io.Writer, shell.Logger) {
-	varsToRedact, short, err := redact.Vars(e.ExecutorConfig.RedactedVars, environ.DumpPairs())
+	varsToRedact, short, err := redact.Vars(e.RedactedVars, environ.DumpPairs())
 	if err != nil {
 		log.OptionalWarningf("bad-redacted-vars", "Couldn't match environment variable names against redacted-vars: %v", err)
 	}
@@ -1217,7 +1219,7 @@ func (e *Executor) setupRedactors(log shell.Logger, environ *env.Environment, st
 	}
 
 	if e.Debug {
-		log.Commentf("Enabling output redaction for values from environment variables matching: %v", e.ExecutorConfig.RedactedVars)
+		log.Commentf("Enabling output redaction for values from environment variables matching: %v", e.RedactedVars)
 	}
 
 	needles := make([]string, 0, len(varsToRedact))
@@ -1225,9 +1227,9 @@ func (e *Executor) setupRedactors(log shell.Logger, environ *env.Environment, st
 		needles = append(needles, pair.Value)
 	}
 
-	stdoutRedactor := replacer.New(stdout, needles, redact.Redact)
+	stdoutRedactor := replacer.New(stdout, needles, redact.Redacted)
 	e.redactors.Append(stdoutRedactor)
-	loggerRedactor := replacer.New(stderr, needles, redact.Redact)
+	loggerRedactor := replacer.New(stderr, needles, redact.Redacted)
 	e.redactors.Append(loggerRedactor)
 
 	logger := shell.NewWriterLogger(loggerRedactor, true, e.DisabledWarnings)
@@ -1297,25 +1299,18 @@ func (e *Executor) kubernetesSetup(ctx context.Context, environ *env.Environment
 		}
 	}
 
-	// Proceed when ready
-	if err := k8sAgentSocket.Await(ctx, kubernetes.RunStateStart); err != nil {
-		return fmt.Errorf("error waiting for client to become ready: %w", err)
-	}
-
-	go func() {
-		// If the k8s client is interrupted because the "server" agent is
-		// stopped or unreachable, we should stop running the job.
-		err := k8sAgentSocket.Await(ctx, kubernetes.RunStateInterrupt)
-		// If the k8s client is interrupted because our own ctx was cancelled,
-		// then the job is already stopping, so there's no point logging an
-		// error.
-		if errors.Is(err, context.Canceled) {
-			return
-		}
+	// So that the agent doesn't exit early thinking the client is lost, we want
+	// to continue talking to the agent container for as long as possible (after
+	// Interrupt). Hence detach the StatusLoop context from cancellation using
+	// [context.WithoutCancel]. The goroutine will exit with the process.
+	// (Why even have a context arg? Testing and possible future value-passing)
+	return k8sAgentSocket.StatusLoop(context.WithoutCancel(ctx), func(err error) {
+		// If the k8s client is interrupted for any reason (either the server
+		// is in state interrupted or the connection died or ...), we should
+		// cancel the job.
 		if err != nil {
 			e.shell.Errorf("Error waiting for client interrupt: %v", err)
 		}
 		e.Cancel()
-	}()
-	return nil
+	})
 }
