@@ -6,14 +6,15 @@
 package logs
 
 import (
-	"bytes"
 	"context"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"math"
 	"math/rand"
 	"net/url"
 	"reflect"
+	"slices"
 	"strings"
 	"sync"
 	"time"
@@ -46,26 +47,29 @@ type Logger interface {
 // the struct. Any changes here MUST be reflected in the AST()
 // implementation below.
 type EventV1 struct {
-	Labels         map[string]string       `json:"labels"`
-	DecisionID     string                  `json:"decision_id"`
-	TraceID        string                  `json:"trace_id,omitempty"`
-	SpanID         string                  `json:"span_id,omitempty"`
-	Revision       string                  `json:"revision,omitempty"` // Deprecated: Use Bundles instead
-	Bundles        map[string]BundleInfoV1 `json:"bundles,omitempty"`
-	Path           string                  `json:"path,omitempty"`
-	Query          string                  `json:"query,omitempty"`
-	Input          *interface{}            `json:"input,omitempty"`
-	Result         *interface{}            `json:"result,omitempty"`
-	MappedResult   *interface{}            `json:"mapped_result,omitempty"`
-	NDBuiltinCache *interface{}            `json:"nd_builtin_cache,omitempty"`
-	Erased         []string                `json:"erased,omitempty"`
-	Masked         []string                `json:"masked,omitempty"`
-	Error          error                   `json:"error,omitempty"`
-	RequestedBy    string                  `json:"requested_by,omitempty"`
-	Timestamp      time.Time               `json:"timestamp"`
-	Metrics        map[string]interface{}  `json:"metrics,omitempty"`
-	RequestID      uint64                  `json:"req_id,omitempty"`
-	RequestContext *RequestContext         `json:"request_context,omitempty"`
+	Labels              map[string]string       `json:"labels"`
+	DecisionID          string                  `json:"decision_id"`
+	BatchDecisionID     string                  `json:"batch_decision_id,omitempty"`
+	TraceID             string                  `json:"trace_id,omitempty"`
+	SpanID              string                  `json:"span_id,omitempty"`
+	Revision            string                  `json:"revision,omitempty"` // Deprecated: Use Bundles instead
+	Bundles             map[string]BundleInfoV1 `json:"bundles,omitempty"`
+	Path                string                  `json:"path,omitempty"`
+	Query               string                  `json:"query,omitempty"`
+	Input               *any                    `json:"input,omitempty"`
+	Result              *any                    `json:"result,omitempty"`
+	IntermediateResults map[string]any          `json:"intermediate_results,omitempty"`
+	MappedResult        *any                    `json:"mapped_result,omitempty"`
+	NDBuiltinCache      *any                    `json:"nd_builtin_cache,omitempty"`
+	Erased              []string                `json:"erased,omitempty"`
+	Masked              []string                `json:"masked,omitempty"`
+	Error               error                   `json:"error,omitempty"`
+	RequestedBy         string                  `json:"requested_by,omitempty"`
+	Timestamp           time.Time               `json:"timestamp"`
+	Metrics             map[string]any          `json:"metrics,omitempty"`
+	RequestID           uint64                  `json:"req_id,omitempty"`
+	RequestContext      *RequestContext         `json:"request_context,omitempty"`
+	Custom              map[string]any          `json:"custom,omitempty"`
 
 	inputAST ast.Value
 }
@@ -87,51 +91,36 @@ type HTTPRequestContext struct {
 func (b *BundleInfoV1) AST() ast.Value {
 	result := ast.NewObject()
 	if len(b.Revision) > 0 {
-		result.Insert(ast.StringTerm("revision"), ast.StringTerm(b.Revision))
+		result.Insert(ast.InternedTerm("revision"), ast.StringTerm(b.Revision))
 	}
 	return result
 }
 
-// Key ast.Term values for the Rego AST representation of the EventV1
-var labelsKey = ast.StringTerm("labels")
-var decisionIDKey = ast.StringTerm("decision_id")
-var revisionKey = ast.StringTerm("revision")
-var bundlesKey = ast.StringTerm("bundles")
-var pathKey = ast.StringTerm("path")
-var queryKey = ast.StringTerm("query")
-var inputKey = ast.StringTerm("input")
-var resultKey = ast.StringTerm("result")
-var mappedResultKey = ast.StringTerm("mapped_result")
-var ndBuiltinCacheKey = ast.StringTerm("nd_builtin_cache")
-var erasedKey = ast.StringTerm("erased")
-var maskedKey = ast.StringTerm("masked")
-var errorKey = ast.StringTerm("error")
-var requestedByKey = ast.StringTerm("requested_by")
-var timestampKey = ast.StringTerm("timestamp")
-var metricsKey = ast.StringTerm("metrics")
-var requestIDKey = ast.StringTerm("req_id")
-
 // AST returns the Rego AST representation for a given EventV1 object.
 // This avoids having to round trip through JSON while applying a decision log
 // mask policy to the event.
 func (e *EventV1) AST() (ast.Value, error) {
 	var err error
-	event := ast.NewObject()
+	event := ast.NewObject(
+		ast.Item(ast.InternedTerm("decision_id"), ast.StringTerm(e.DecisionID)),
+	)
+
+	if e.BatchDecisionID != "" {
+		event.Insert(ast.InternedTerm("batch_decision_id"), ast.StringTerm(e.BatchDecisionID))
+	}
 
 	if e.Labels != nil {
 		labelsObj := ast.NewObject()
 		for k, v := range e.Labels {
 			labelsObj.Insert(ast.StringTerm(k), ast.StringTerm(v))
 		}
-		event.Insert(labelsKey, ast.NewTerm(labelsObj))
+		event.Insert(ast.InternedTerm("labels"), ast.NewTerm(labelsObj))
 	} else {
-		event.Insert(labelsKey, ast.NullTerm())
+		event.Insert(ast.InternedTerm("labels"), ast.NullTerm())
 	}
 
-	event.Insert(decisionIDKey, ast.StringTerm(e.DecisionID))
-
 	if len(e.Revision) > 0 {
-		event.Insert(revisionKey, ast.StringTerm(e.Revision))
+		event.Insert(ast.InternedTerm("revision"), ast.StringTerm(e.Revision))
 	}
 
 	if len(e.Bundles) > 0 {
@@ -139,25 +128,25 @@ func (e *EventV1) AST() (ast.Value, error) {
 		for k, v := range e.Bundles {
 			bundlesObj.Insert(ast.StringTerm(k), ast.NewTerm(v.AST()))
 		}
-		event.Insert(bundlesKey, ast.NewTerm(bundlesObj))
+		event.Insert(ast.InternedTerm("bundles"), ast.NewTerm(bundlesObj))
 	}
 
 	if len(e.Path) > 0 {
-		event.Insert(pathKey, ast.StringTerm(e.Path))
+		event.Insert(ast.InternedTerm("path"), ast.StringTerm(e.Path))
 	}
 
 	if len(e.Query) > 0 {
-		event.Insert(queryKey, ast.StringTerm(e.Query))
+		event.Insert(ast.InternedTerm("query"), ast.StringTerm(e.Query))
 	}
 
-	if e.Input != nil {
-		if e.inputAST == nil {
-			e.inputAST, err = roundtripJSONToAST(e.Input)
-			if err != nil {
-				return nil, err
-			}
+	if e.inputAST != nil {
+		event.Insert(ast.InternedTerm("input"), ast.NewTerm(e.inputAST))
+	} else if e.Input != nil {
+		e.inputAST, err = roundtripJSONToAST(e.Input)
+		if err != nil {
+			return nil, err
 		}
-		event.Insert(inputKey, ast.NewTerm(e.inputAST))
+		event.Insert(ast.InternedTerm("input"), ast.NewTerm(e.inputAST))
 	}
 
 	if e.Result != nil {
@@ -165,7 +154,15 @@ func (e *EventV1) AST() (ast.Value, error) {
 		if err != nil {
 			return nil, err
 		}
-		event.Insert(resultKey, ast.NewTerm(results))
+		event.Insert(ast.InternedTerm("result"), ast.NewTerm(results))
+	}
+
+	if len(e.IntermediateResults) > 0 {
+		iresults, err := roundtripJSONToAST(e.IntermediateResults)
+		if err != nil {
+			return nil, err
+		}
+		event.Insert(ast.InternedTerm("intermediate_results"), ast.NewTerm(iresults))
 	}
 
 	if e.MappedResult != nil {
@@ -173,7 +170,7 @@ func (e *EventV1) AST() (ast.Value, error) {
 		if err != nil {
 			return nil, err
 		}
-		event.Insert(mappedResultKey, ast.NewTerm(mResults))
+		event.Insert(ast.InternedTerm("mapped_result"), ast.NewTerm(mResults))
 	}
 
 	if e.NDBuiltinCache != nil {
@@ -181,7 +178,7 @@ func (e *EventV1) AST() (ast.Value, error) {
 		if err != nil {
 			return nil, err
 		}
-		event.Insert(ndBuiltinCacheKey, ast.NewTerm(ndbCache))
+		event.Insert(ast.InternedTerm("nd_builtin_cache"), ast.NewTerm(ndbCache))
 	}
 
 	if len(e.Erased) > 0 {
@@ -189,7 +186,7 @@ func (e *EventV1) AST() (ast.Value, error) {
 		for i, v := range e.Erased {
 			erased[i] = ast.StringTerm(v)
 		}
-		event.Insert(erasedKey, ast.NewTerm(ast.NewArray(erased...)))
+		event.Insert(ast.InternedTerm("erased"), ast.ArrayTerm(erased...))
 	}
 
 	if len(e.Masked) > 0 {
@@ -197,7 +194,7 @@ func (e *EventV1) AST() (ast.Value, error) {
 		for i, v := range e.Masked {
 			masked[i] = ast.StringTerm(v)
 		}
-		event.Insert(maskedKey, ast.NewTerm(ast.NewArray(masked...)))
+		event.Insert(ast.InternedTerm("masked"), ast.ArrayTerm(masked...))
 	}
 
 	if e.Error != nil {
@@ -205,11 +202,11 @@ func (e *EventV1) AST() (ast.Value, error) {
 		if err != nil {
 			return nil, err
 		}
-		event.Insert(errorKey, ast.NewTerm(evalErr))
+		event.Insert(ast.InternedTerm("error"), ast.NewTerm(evalErr))
 	}
 
 	if len(e.RequestedBy) > 0 {
-		event.Insert(requestedByKey, ast.StringTerm(e.RequestedBy))
+		event.Insert(ast.InternedTerm("requested_by"), ast.StringTerm(e.RequestedBy))
 	}
 
 	// Use the timestamp JSON marshaller to ensure the format is the same as
@@ -218,27 +215,35 @@ func (e *EventV1) AST() (ast.Value, error) {
 	if err != nil {
 		return nil, err
 	}
-	event.Insert(timestampKey, ast.StringTerm(strings.Trim(string(timeBytes), "\"")))
+	event.Insert(ast.InternedTerm("timestamp"), ast.StringTerm(strings.Trim(string(timeBytes), "\"")))
 
 	if e.Metrics != nil {
 		m, err := ast.InterfaceToValue(e.Metrics)
 		if err != nil {
 			return nil, err
 		}
-		event.Insert(metricsKey, ast.NewTerm(m))
+		event.Insert(ast.InternedTerm("metrics"), ast.NewTerm(m))
 	}
 
 	if e.RequestID > 0 {
-		event.Insert(requestIDKey, ast.UIntNumberTerm(e.RequestID))
+		event.Insert(ast.InternedTerm("req_id"), ast.UIntNumberTerm(e.RequestID))
+	}
+
+	if len(e.Custom) > 0 {
+		custom, err := roundtripJSONToAST(e.Custom)
+		if err != nil {
+			return nil, err
+		}
+		event.Insert(ast.InternedTerm("custom"), ast.NewTerm(custom))
 	}
 
 	return event, nil
 }
 
-func roundtripJSONToAST(x interface{}) (ast.Value, error) {
+func roundtripJSONToAST(x any) (ast.Value, error) {
 	rawPtr := util.Reference(x)
 	// roundtrip through json: this turns slices (e.g. []string, []bool) into
-	// []interface{}, the only array type ast.InterfaceToValue can work with
+	// []any, the only array type ast.InterfaceToValue can work with
 	if err := util.RoundTrip(rawPtr); err != nil {
 		return nil, err
 	}
@@ -251,20 +256,27 @@ const (
 	minRetryDelay                       = time.Millisecond * 100
 	defaultMinDelaySeconds              = int64(300)
 	defaultMaxDelaySeconds              = int64(600)
-	defaultUploadSizeLimitBytes         = int64(32768) // 32KB limit
-	defaultBufferSizeLimitBytes         = int64(0)     // unlimited
+	defaultBufferSizeLimitEvents        = int64(10000)
+	defaultUploadSizeLimitBytes         = int64(32768)      // 32KB limit
+	minUploadSizeLimitBytes             = int64(90)         // A single event with a decision ID (69 bytes) + empty gzip file (21 bytes)
+	maxUploadSizeLimitBytes             = int64(4294967296) // about 4GB
+	defaultBufferSizeLimitBytes         = int64(0)          // unlimited
 	defaultMaskDecisionPath             = "/system/log/mask"
 	defaultDropDecisionPath             = "/system/log/drop"
 	logRateLimitExDropCounterName       = "decision_logs_dropped_rate_limit_exceeded"
-	logNDBDropCounterName               = "decision_logs_nd_builtin_cache_dropped"
+	logBufferEventDropCounterName       = "decision_logs_dropped_buffer_size_limit_exceeded"
 	logBufferSizeLimitExDropCounterName = "decision_logs_dropped_buffer_size_limit_bytes_exceeded"
 	logEncodingFailureCounterName       = "decision_logs_encoding_failure"
 	defaultResourcePath                 = "/logs"
+	sizeBufferType                      = "size"
+	eventBufferType                     = "event"
 )
 
 // ReportingConfig represents configuration for the plugin's reporting behaviour.
 type ReportingConfig struct {
-	BufferSizeLimitBytes  *int64               `json:"buffer_size_limit_bytes,omitempty"`  // max size of in-memory buffer
+	BufferType            string               `json:"buffer_type,omitempty"`              // toggles how the buffer stores events, defaults to using bytes
+	BufferSizeLimitBytes  *int64               `json:"buffer_size_limit_bytes,omitempty"`  // max size of in-memory size buffer
+	BufferSizeLimitEvents *int64               `json:"buffer_size_limit_events,omitempty"` // max size of in-memory event channel buffer
 	UploadSizeLimitBytes  *int64               `json:"upload_size_limit_bytes,omitempty"`  // max size of upload payload
 	MinDelaySeconds       *int64               `json:"min_delay_seconds,omitempty"`        // min amount of time to wait between successful poll attempts
 	MaxDelaySeconds       *int64               `json:"max_delay_seconds,omitempty"`        // max amount of time to wait between poll attempts
@@ -296,15 +308,12 @@ type Config struct {
 	dropDecisionRef ast.Ref
 }
 
-func (c *Config) validateAndInjectDefaults(services []string, pluginsList []string, trigger *plugins.TriggerMode) error {
+func (c *Config) validateAndInjectDefaults(services []string, pluginsList []string, trigger *plugins.TriggerMode, l logging.Logger) error {
 
 	if c.Plugin != nil {
 		var found bool
-		for _, other := range pluginsList {
-			if other == *c.Plugin {
-				found = true
-				break
-			}
+		if slices.Contains(pluginsList, *c.Plugin) {
+			found = true
 		}
 		if !found {
 			return fmt.Errorf("invalid plugin name %q in decision_logs", *c.Plugin)
@@ -316,14 +325,7 @@ func (c *Config) validateAndInjectDefaults(services []string, pluginsList []stri
 		// both console logs and the default service option.
 		c.Service = services[0]
 	} else if c.Service != "" {
-		found := false
-
-		for _, svc := range services {
-			if svc == c.Service {
-				found = true
-				break
-			}
-		}
+		found := slices.Contains(services, c.Service)
 
 		if !found {
 			return fmt.Errorf("invalid service name %q in decision_logs", c.Service)
@@ -342,14 +344,14 @@ func (c *Config) validateAndInjectDefaults(services []string, pluginsList []stri
 	// reject bad min/max values
 	if c.Reporting.MaxDelaySeconds != nil && c.Reporting.MinDelaySeconds != nil {
 		if *c.Reporting.MaxDelaySeconds < *c.Reporting.MinDelaySeconds {
-			return fmt.Errorf("max reporting delay must be >= min reporting delay in decision_logs")
+			return errors.New("max reporting delay must be >= min reporting delay in decision_logs")
 		}
 		min = *c.Reporting.MinDelaySeconds
 		max = *c.Reporting.MaxDelaySeconds
 	} else if c.Reporting.MaxDelaySeconds == nil && c.Reporting.MinDelaySeconds != nil {
-		return fmt.Errorf("reporting configuration missing 'max_delay_seconds' in decision_logs")
+		return errors.New("reporting configuration missing 'max_delay_seconds' in decision_logs")
 	} else if c.Reporting.MinDelaySeconds == nil && c.Reporting.MaxDelaySeconds != nil {
-		return fmt.Errorf("reporting configuration missing 'min_delay_seconds' in decision_logs")
+		return errors.New("reporting configuration missing 'min_delay_seconds' in decision_logs")
 	}
 
 	// scale to seconds
@@ -365,19 +367,58 @@ func (c *Config) validateAndInjectDefaults(services []string, pluginsList []stri
 		uploadLimit = *c.Reporting.UploadSizeLimitBytes
 	}
 
-	c.Reporting.UploadSizeLimitBytes = &uploadLimit
+	switch {
+	case uploadLimit > maxUploadSizeLimitBytes:
+		maxUploadLimit := maxUploadSizeLimitBytes
+		c.Reporting.UploadSizeLimitBytes = &maxUploadLimit
+		if l != nil {
+			l.Warn("the configured `upload_size_limit_bytes` (%d) has been set to the maximum limit (%d)", uploadLimit, maxUploadLimit)
+		}
+	case uploadLimit < minUploadSizeLimitBytes:
+		minUploadLimit := minUploadSizeLimitBytes
+		c.Reporting.UploadSizeLimitBytes = &minUploadLimit
+		if l != nil {
+			l.Warn("the configured `upload_size_limit_bytes` (%d) has been set to the minimum limit (%d)", uploadLimit, minUploadLimit)
+		}
+	default:
+		c.Reporting.UploadSizeLimitBytes = &uploadLimit
+	}
+
+	if c.Reporting.BufferType == "" {
+		c.Reporting.BufferType = sizeBufferType
+	} else if c.Reporting.BufferType != eventBufferType && c.Reporting.BufferType != sizeBufferType {
+		return fmt.Errorf("invalid buffer type %q, expected %q or %q", c.Reporting.BufferType, eventBufferType, sizeBufferType)
+	}
+
+	if c.Reporting.BufferType == eventBufferType && c.Reporting.BufferSizeLimitBytes != nil {
+		return fmt.Errorf("invalid decision_log config, 'buffer_size_limit_bytes' isn't supported for the %v buffer type", eventBufferType)
+	}
+	if c.Reporting.BufferType == sizeBufferType && c.Reporting.BufferSizeLimitEvents != nil {
+		return fmt.Errorf("invalid decision_log config, 'buffer_size_limit_events' isn't supported for the %v buffer type", sizeBufferType)
+	}
 
 	if c.Reporting.BufferSizeLimitBytes != nil && c.Reporting.MaxDecisionsPerSecond != nil {
-		return fmt.Errorf("invalid decision_log config, specify either 'buffer_size_limit_bytes' or 'max_decisions_per_second'")
+		return errors.New("invalid decision_log config, specify either 'buffer_size_limit_bytes' or 'max_decisions_per_second'")
 	}
 
 	// default the buffer size limit
-	bufferLimit := defaultBufferSizeLimitBytes
+	sizeBufferLimit := defaultBufferSizeLimitBytes
 	if c.Reporting.BufferSizeLimitBytes != nil {
-		bufferLimit = *c.Reporting.BufferSizeLimitBytes
+		if *c.Reporting.BufferSizeLimitBytes <= int64(0) {
+			return errors.New("invalid decision_log config, 'buffer_size_limit_bytes' must be higher than 0")
+		}
+		sizeBufferLimit = *c.Reporting.BufferSizeLimitBytes
 	}
+	c.Reporting.BufferSizeLimitBytes = &sizeBufferLimit
 
-	c.Reporting.BufferSizeLimitBytes = &bufferLimit
+	eventBufferLimit := defaultBufferSizeLimitEvents
+	if c.Reporting.BufferSizeLimitEvents != nil {
+		if *c.Reporting.BufferSizeLimitEvents <= int64(0) {
+			return errors.New("invalid decision_log config, 'buffer_size_limit_entries' must be higher than 0")
+		}
+		eventBufferLimit = *c.Reporting.BufferSizeLimitEvents
+	}
+	c.Reporting.BufferSizeLimitEvents = &eventBufferLimit
 
 	if c.MaskDecision == nil {
 		maskDecision := defaultMaskDecisionPath
@@ -416,20 +457,23 @@ func (c *Config) validateAndInjectDefaults(services []string, pluginsList []stri
 
 // Plugin implements decision log buffering and uploading.
 type Plugin struct {
-	manager      *plugins.Manager
-	config       Config
-	buffer       *logBuffer
-	enc          *chunkEncoder
-	mtx          sync.Mutex
-	statusMtx    sync.Mutex
-	stop         chan chan struct{}
-	reconfig     chan reconfigure
-	preparedMask prepareOnce
-	preparedDrop prepareOnce
-	limiter      *rate.Limiter
-	metrics      metrics.Metrics
-	logger       logging.Logger
-	status       *lstat.Status
+	manager       *plugins.Manager
+	config        Config
+	runningBuffer string
+	reconfigMtx   sync.RWMutex // reconfigMtx blocks reads/writes on buffer reconfiguration
+	eventBuffer   *eventBuffer
+	buffer        *logBuffer
+	enc           *chunkEncoder
+	mtx           sync.Mutex
+	statusMtx     sync.Mutex
+	stop          chan chan struct{}
+	reconfig      chan reconfigure
+	preparedMask  prepareOnce
+	preparedDrop  prepareOnce
+	limiter       *rate.Limiter
+	metrics       metrics.Metrics
+	logger        logging.Logger
+	status        *lstat.Status
 }
 
 type prepareOnce struct {
@@ -456,7 +500,7 @@ func (po *prepareOnce) prepareOnce(f func() (*rego.PreparedEvalQuery, error)) (*
 }
 
 type reconfigure struct {
-	config interface{}
+	config any
 	done   chan struct{}
 }
 
@@ -477,6 +521,7 @@ type ConfigBuilder struct {
 	services []string
 	plugins  []string
 	trigger  *plugins.TriggerMode
+	logger   logging.Logger
 }
 
 // NewConfigBuilder returns a new ConfigBuilder to build and parse the plugin config.
@@ -484,6 +529,11 @@ func NewConfigBuilder() *ConfigBuilder {
 	return &ConfigBuilder{}
 }
 
+func (b *ConfigBuilder) WithLogger(l logging.Logger) *ConfigBuilder {
+	b.logger = l
+	return b
+}
+
 // WithBytes sets the raw plugin config.
 func (b *ConfigBuilder) WithBytes(config []byte) *ConfigBuilder {
 	b.raw = config
@@ -525,7 +575,7 @@ func (b *ConfigBuilder) Parse() (*Config, error) {
 		return nil, nil
 	}
 
-	if err := parsedConfig.validateAndInjectDefaults(b.services, b.plugins, b.trigger); err != nil {
+	if err := parsedConfig.validateAndInjectDefaults(b.services, b.plugins, b.trigger, b.logger); err != nil {
 		return nil, err
 	}
 
@@ -539,15 +589,30 @@ func New(parsedConfig *Config, manager *plugins.Manager) *Plugin {
 		manager:      manager,
 		config:       *parsedConfig,
 		stop:         make(chan chan struct{}),
-		buffer:       newLogBuffer(*parsedConfig.Reporting.BufferSizeLimitBytes),
 		enc:          newChunkEncoder(*parsedConfig.Reporting.UploadSizeLimitBytes),
 		reconfig:     make(chan reconfigure),
-		logger:       manager.Logger().WithFields(map[string]interface{}{"plugin": Name}),
+		logger:       manager.Logger().WithFields(map[string]any{"plugin": Name}),
 		status:       &lstat.Status{},
 		preparedDrop: *newPrepareOnce(),
 		preparedMask: *newPrepareOnce(),
 	}
 
+	plugin.enc.WithLogger(plugin.logger)
+
+	switch parsedConfig.Reporting.BufferType {
+	case eventBufferType:
+		plugin.eventBuffer = newEventBuffer(
+			*parsedConfig.Reporting.BufferSizeLimitEvents,
+			plugin.manager.Client(plugin.config.Service),
+			*parsedConfig.Resource,
+			*parsedConfig.Reporting.UploadSizeLimitBytes,
+		).WithLogger(plugin.logger).WithMetrics(plugin.metrics)
+		plugin.runningBuffer = eventBufferType
+	case sizeBufferType:
+		plugin.buffer = newLogBuffer(*parsedConfig.Reporting.BufferSizeLimitBytes)
+		plugin.runningBuffer = sizeBufferType
+	}
+
 	if parsedConfig.Reporting.MaxDecisionsPerSecond != nil {
 		limit := *parsedConfig.Reporting.MaxDecisionsPerSecond
 		plugin.limiter = rate.NewLimiter(rate.Limit(limit), int(math.Max(1, limit)))
@@ -614,7 +679,7 @@ func (p *Plugin) flushDecisions(ctx context.Context) {
 
 	go func(ctx context.Context, done chan bool) {
 		for ctx.Err() == nil {
-			if _, err := p.oneShot(ctx); err != nil {
+			if err := p.oneShot(ctx); err != nil && !errors.Is(err, &bufferEmpty{}) {
 				p.logger.Error("Error flushing decisions: %s", err)
 				// Wait some before retrying, but skip incrementing interval since we are shutting down
 				time.Sleep(1 * time.Second)
@@ -638,29 +703,31 @@ func (p *Plugin) flushDecisions(ctx context.Context) {
 
 // Log appends a decision log event to the buffer for uploading.
 func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
-
 	bundles := map[string]BundleInfoV1{}
 	for name, info := range decision.Bundles {
 		bundles[name] = BundleInfoV1{Revision: info.Revision}
 	}
 
 	event := EventV1{
-		Labels:         p.manager.Labels(),
-		DecisionID:     decision.DecisionID,
-		TraceID:        decision.TraceID,
-		SpanID:         decision.SpanID,
-		Revision:       decision.Revision,
-		Bundles:        bundles,
-		Path:           decision.Path,
-		Query:          decision.Query,
-		Input:          decision.Input,
-		Result:         decision.Results,
-		MappedResult:   decision.MappedResults,
-		NDBuiltinCache: decision.NDBuiltinCache,
-		RequestedBy:    decision.RemoteAddr,
-		Timestamp:      decision.Timestamp,
-		RequestID:      decision.RequestID,
-		inputAST:       decision.InputAST,
+		Labels:              p.manager.Labels(),
+		DecisionID:          decision.DecisionID,
+		BatchDecisionID:     decision.BatchDecisionID,
+		TraceID:             decision.TraceID,
+		SpanID:              decision.SpanID,
+		Revision:            decision.Revision,
+		Bundles:             bundles,
+		Path:                decision.Path,
+		Query:               decision.Query,
+		Input:               decision.Input,
+		Result:              decision.Results,
+		IntermediateResults: decision.IntermediateResults,
+		MappedResult:        decision.MappedResults,
+		NDBuiltinCache:      decision.NDBuiltinCache,
+		RequestedBy:         decision.RemoteAddr,
+		Timestamp:           decision.Timestamp,
+		RequestID:           decision.RequestID,
+		inputAST:            decision.InputAST,
+		Custom:              decision.Custom,
 	}
 
 	headers := map[string][]string{}
@@ -722,7 +789,7 @@ func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
 	if p.config.Plugin != nil {
 		proxy, ok := p.manager.Plugin(*p.config.Plugin).(Logger)
 		if !ok {
-			return fmt.Errorf("plugin does not implement Logger interface")
+			return errors.New("plugin does not implement Logger interface")
 		}
 		return proxy.Log(ctx, event)
 	}
@@ -731,7 +798,7 @@ func (p *Plugin) Log(ctx context.Context, decision *server.Info) error {
 }
 
 // Reconfigure notifies the plugin with a new configuration.
-func (p *Plugin) Reconfigure(_ context.Context, config interface{}) {
+func (p *Plugin) Reconfigure(_ context.Context, config any) {
 
 	done := make(chan struct{})
 	p.reconfig <- reconfigure{config: config, done: done}
@@ -776,13 +843,11 @@ func (p *Plugin) compilerUpdated(storage.Transaction) {
 }
 
 func (p *Plugin) loop() {
-
 	ctx, cancel := context.WithCancel(context.Background())
 
 	var retry int
 
 	for {
-
 		var waitC chan struct{}
 
 		if *p.config.Reporting.Trigger == plugins.TriggerPeriodic && p.config.Service != "" {
@@ -820,7 +885,7 @@ func (p *Plugin) loop() {
 		select {
 		case <-waitC:
 		case update := <-p.reconfig:
-			p.reconfigure(update.config)
+			p.reconfigure(ctx, update.config)
 			update.done <- struct{}{}
 		case done := <-p.stop:
 			cancel()
@@ -830,31 +895,36 @@ func (p *Plugin) loop() {
 	}
 }
 
+type bufferEmpty struct{}
+
+func (*bufferEmpty) Error() string {
+	return "buffer is empty"
+}
+
 func (p *Plugin) doOneShot(ctx context.Context) error {
-	uploaded, err := p.oneShot(ctx)
-
-	// Make a local copy of the plugins's status.
-	p.statusMtx.Lock()
-	p.status.SetError(err)
-	oldStatus := p.status
-	p.statusMtx.Unlock()
-
-	if s := status.Lookup(p.manager); s != nil {
-		s.UpdateDecisionLogsStatus(*oldStatus)
-	}
+	err := p.oneShot(ctx)
 
 	if err != nil {
-		p.logger.Error("%v.", err)
-	} else if uploaded {
-		p.logger.Info("Logs uploaded successfully.")
+		if errors.Is(err, &bufferEmpty{}) {
+			p.logger.Debug("Log upload queue was empty.")
+			err = nil
+		} else {
+			p.logger.Error("%v.", err)
+		}
 	} else {
-		p.logger.Debug("Log upload queue was empty.")
+		p.logger.Info("Logs uploaded successfully.")
 	}
+
+	p.setStatus(err)
 	return err
 }
 
-func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
-	// Make a local copy of the plugins's encoder and buffer and create
+func (p *Plugin) oneShot(ctx context.Context) error {
+	if p.runningBuffer == eventBufferType {
+		return p.eventBuffer.Upload(ctx)
+	}
+
+	// Make a local copy of the plugin's encoder and buffer and create
 	// a new encoder and buffer. This is needed as locking the buffer for
 	// the upload duration will block policy evaluation and result in
 	// increased latency for OPA clients
@@ -862,7 +932,8 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 	oldChunkEnc := p.enc
 	oldBuffer := p.buffer
 	p.buffer = newLogBuffer(*p.config.Reporting.BufferSizeLimitBytes)
-	p.enc = newChunkEncoder(*p.config.Reporting.UploadSizeLimitBytes).WithMetrics(p.metrics)
+	p.enc = newChunkEncoder(*p.config.Reporting.UploadSizeLimitBytes).WithMetrics(p.metrics).WithLogger(p.logger).
+		WithUncompressedLimit(oldChunkEnc.uncompressedLimit, oldChunkEnc.uncompressedLimitScaleDownExponent, oldChunkEnc.uncompressedLimitScaleUpExponent)
 	p.mtx.Unlock()
 
 	// Along with uploading the compressed events in the buffer
@@ -870,7 +941,7 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 	// underlying writer and add to the buffer.
 	chunk, err := oldChunkEnc.Flush()
 	if err != nil {
-		return false, err
+		return err
 	}
 
 	for _, ch := range chunk {
@@ -878,7 +949,7 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 	}
 
 	if oldBuffer.Len() == 0 {
-		return false, nil
+		return &bufferEmpty{}
 	}
 
 	for bs := oldBuffer.Pop(); bs != nil; bs = oldBuffer.Pop() {
@@ -892,8 +963,8 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 					continue
 				}
 
-				for _, event := range events {
-					p.encodeAndBufferEvent(event)
+				for i := range events {
+					p.encodeAndBufferEvent(events[i])
 				}
 			} else {
 				// requeue the chunk
@@ -904,11 +975,10 @@ func (p *Plugin) oneShot(ctx context.Context) (ok bool, err error) {
 		}
 	}
 
-	return err == nil, err
+	return err
 }
 
-func (p *Plugin) reconfigure(config interface{}) {
-
+func (p *Plugin) reconfigure(ctx context.Context, config any) {
 	newConfig := config.(*Config)
 
 	if reflect.DeepEqual(p.config, *newConfig) {
@@ -918,81 +988,89 @@ func (p *Plugin) reconfigure(config interface{}) {
 
 	p.logger.Info("Decision log uploader configuration changed.")
 	p.config = *newConfig
+
+	p.reconfigMtx.Lock()
+	defer p.reconfigMtx.Unlock()
+
+	switch newConfig.Reporting.BufferType {
+	case eventBufferType:
+		if p.eventBuffer == nil {
+			p.eventBuffer = newEventBuffer(
+				*p.config.Reporting.BufferSizeLimitEvents,
+				p.manager.Client(p.config.Service),
+				*p.config.Resource,
+				*p.config.Reporting.UploadSizeLimitBytes).WithLogger(p.logger).WithMetrics(p.metrics)
+		} else {
+			p.eventBuffer.Reconfigure(
+				*p.config.Reporting.BufferSizeLimitEvents,
+				p.manager.Client(p.config.Service),
+				*p.config.Resource,
+				*p.config.Reporting.UploadSizeLimitBytes)
+		}
+
+		if p.runningBuffer == sizeBufferType {
+			if err := p.oneShot(ctx); err != nil && !errors.Is(err, &bufferEmpty{}) {
+				p.setStatus(err)
+			}
+		}
+
+		p.runningBuffer = eventBufferType
+	case sizeBufferType:
+		if p.runningBuffer == eventBufferType {
+			if err := p.eventBuffer.Upload(ctx); err != nil && !errors.Is(err, &bufferEmpty{}) {
+				p.setStatus(err)
+			}
+		}
+
+		if p.buffer == nil {
+			p.buffer = newLogBuffer(*p.config.Reporting.BufferSizeLimitBytes)
+		}
+
+		p.runningBuffer = sizeBufferType
+	}
 }
 
 // NOTE(philipc): Because ND builtins caching can cause unbounded growth in
 // decision log entry size, we do best-effort event encoding here, and when we
 // run out of space, we drop the ND builtins cache, and try encoding again.
 func (p *Plugin) encodeAndBufferEvent(event EventV1) {
-	if p.limiter != nil {
-		if !p.limiter.Allow() {
-			if p.metrics != nil {
-				p.metrics.Counter(logRateLimitExDropCounterName).Incr()
-			}
-
-			p.logger.Error("Decision log dropped as rate limit exceeded. Reduce reporting interval or increase rate limit.")
-			return
-		}
+	if p.limiter != nil && !p.limiter.Allow() {
+		p.incrMetric(logRateLimitExDropCounterName)
+		p.logger.Error("Decision log dropped as rate limit exceeded. Reduce reporting interval or increase rate limit.")
+		return
 	}
 
-	result, err := p.encodeEvent(event)
+	// only blocks when the buffer is being reconfigured
+	p.reconfigMtx.RLock()
+	defer p.reconfigMtx.RUnlock()
+
+	if p.runningBuffer == eventBufferType {
+		p.eventBuffer.Push(&event)
+		return
+	}
+
+	eventBytes, err := json.Marshal(&event)
 	if err != nil {
-		// If there's no ND builtins cache in the event, then we don't
-		// need to retry encoding anything.
-		if event.NDBuiltinCache == nil {
-			// TODO(tsandall): revisit this now that we have an API that
-			// can return an error. Should the default behaviour be to
-			// fail-closed as we do for plugins?
-
-			if p.metrics != nil {
-				p.metrics.Counter(logEncodingFailureCounterName).Incr()
-			}
-			p.logger.Error("Log encoding failed: %v.", err)
-			return
-		}
-
-		// Attempt to encode the event again, dropping the ND builtins cache.
-		newEvent := event
-		newEvent.NDBuiltinCache = nil
-
-		result, err = p.encodeEvent(newEvent)
-		if err != nil {
-			if p.metrics != nil {
-				p.metrics.Counter(logEncodingFailureCounterName).Incr()
-			}
-			p.logger.Error("Log encoding failed: %v.", err)
-			return
-		}
-
-		// Re-encoding was successful, but we still need to alert users.
-		p.logger.Error("ND builtins cache dropped from this event to fit under maximum upload size limits. Increase upload size limit or change usage of non-deterministic builtins.")
-		p.metrics.Counter(logNDBDropCounterName).Incr()
+		p.logger.Error("Decision log dropped due to error serializing event to JSON: %v", err)
+		return
 	}
 
 	p.mtx.Lock()
 	defer p.mtx.Unlock()
+	result, err := p.enc.Encode(event, eventBytes)
+	if err != nil {
+		return
+	}
 	for _, chunk := range result {
 		p.bufferChunk(p.buffer, chunk)
 	}
 }
 
-func (p *Plugin) encodeEvent(event EventV1) ([][]byte, error) {
-	var buf bytes.Buffer
-	if err := json.NewEncoder(&buf).Encode(event); err != nil {
-		return nil, err
-	}
-
-	p.mtx.Lock()
-	defer p.mtx.Unlock()
-	return p.enc.WriteBytes(buf.Bytes())
-}
-
 func (p *Plugin) bufferChunk(buffer *logBuffer, bs []byte) {
 	dropped := buffer.Push(bs)
 	if dropped > 0 {
-		if p.metrics != nil {
-			p.metrics.Counter(logBufferSizeLimitExDropCounterName).Incr()
-		}
+		p.incrMetric(logBufferEventDropCounterName)
+		p.incrMetric(logBufferSizeLimitExDropCounterName)
 		p.logger.Error("Dropped %v chunks from buffer. Reduce reporting interval or increase buffer size.", dropped)
 	}
 }
@@ -1118,13 +1196,30 @@ func (p *Plugin) logEvent(event EventV1) error {
 	if err != nil {
 		return err
 	}
-	fields := map[string]interface{}{}
+	fields := map[string]any{}
 	err = util.UnmarshalJSON(eventBuf, &fields)
 	if err != nil {
 		return err
 	}
-	p.manager.ConsoleLogger().WithFields(fields).WithFields(map[string]interface{}{
+	p.manager.ConsoleLogger().WithFields(fields).WithFields(map[string]any{
 		"type": "openpolicyagent.org/decision_logs",
 	}).Info("Decision Log")
 	return nil
 }
+
+func (p *Plugin) incrMetric(name string) {
+	if p.metrics != nil {
+		p.metrics.Counter(name).Incr()
+	}
+}
+
+func (p *Plugin) setStatus(err error) {
+	p.statusMtx.Lock()
+	p.status.SetError(err)
+	oldStatus := p.status
+	p.statusMtx.Unlock()
+
+	if s := status.Lookup(p.manager); s != nil {
+		s.UpdateDecisionLogsStatus(*oldStatus)
+	}
+}
