@@ -62,6 +62,7 @@ type Query struct {
 	printHook                   print.Hook
 	tracingOpts                 tracing.Options
 	virtualCache                VirtualCache
+	baseCache                   BaseCache
 }
 
 // Builtin represents a built-in function that queries can call.
@@ -77,7 +78,6 @@ func NewQuery(query ast.Body) *Query {
 		genvarprefix: ast.WildcardPrefix,
 		indexing:     true,
 		earlyExit:    true,
-		external:     newResolverTrie(),
 	}
 }
 
@@ -277,6 +277,9 @@ func (q *Query) WithBuiltinErrorList(list *[]Error) *Query {
 
 // WithResolver configures an external resolver to use for the given ref.
 func (q *Query) WithResolver(ref ast.Ref, r resolver.Resolver) *Query {
+	if q.external == nil {
+		q.external = newResolverTrie()
+	}
 	q.external.Put(ref, r)
 	return q
 }
@@ -314,6 +317,13 @@ func (q *Query) WithVirtualCache(vc VirtualCache) *Query {
 	return q
 }
 
+// WithBaseCache sets the BaseCache to use during evaluation. This is
+// optional, and if not set, the default cache is used.
+func (q *Query) WithBaseCache(bc BaseCache) *Query {
+	q.baseCache = bc
+	return q
+}
+
 // WithNondeterministicBuiltins causes non-deterministic builtins to be evalued
 // during partial evaluation. This is needed to pull in external data, or validate
 // a JWT, during PE, so that the result informs what queries are returned.
@@ -336,7 +346,7 @@ func (q *Query) PartialRun(ctx context.Context) (partials []ast.Body, support []
 	if q.seed == nil {
 		q.seed = rand.Reader
 	}
-	if !q.time.IsZero() {
+	if q.time.IsZero() {
 		q.time = time.Now()
 	}
 	if q.metrics == nil {
@@ -353,6 +363,13 @@ func (q *Query) PartialRun(ctx context.Context) (partials []ast.Body, support []
 		vc = NewVirtualCache()
 	}
 
+	var bc BaseCache
+	if q.baseCache != nil {
+		bc = q.baseCache
+	} else {
+		bc = newBaseCache()
+	}
+
 	e := &eval{
 		ctx:                         ctx,
 		metrics:                     q.metrics,
@@ -366,8 +383,7 @@ func (q *Query) PartialRun(ctx context.Context) (partials []ast.Body, support []
 		bindings:                    b,
 		compiler:                    q.compiler,
 		store:                       q.store,
-		baseCache:                   newBaseCache(),
-		targetStack:                 newRefStack(),
+		baseCache:                   bc,
 		txn:                         q.txn,
 		input:                       q.input,
 		external:                    q.external,
@@ -377,16 +393,14 @@ func (q *Query) PartialRun(ctx context.Context) (partials []ast.Body, support []
 		instr:                       q.instr,
 		builtins:                    q.builtins,
 		builtinCache:                builtins.Cache{},
-		functionMocks:               newFunctionMocksStack(),
 		interQueryBuiltinCache:      q.interQueryBuiltinCache,
 		interQueryBuiltinValueCache: q.interQueryBuiltinValueCache,
 		ndBuiltinCache:              q.ndBuiltinCache,
 		virtualCache:                vc,
-		comprehensionCache:          newComprehensionCache(),
 		saveSet:                     newSaveSet(q.unknowns, b, q.instr),
 		saveStack:                   newSaveStack(),
 		saveSupport:                 newSaveSupport(),
-		saveNamespace:               ast.StringTerm(q.partialNamespace),
+		saveNamespace:               ast.InternedTerm(q.partialNamespace),
 		skipSaveNamespace:           q.skipSaveNamespace,
 		inliningControl: &inliningControl{
 			shallow:                  q.shallowInlining,
@@ -544,6 +558,13 @@ func (q *Query) Iter(ctx context.Context, iter func(QueryResult) error) error {
 		vc = NewVirtualCache()
 	}
 
+	var bc BaseCache
+	if q.baseCache != nil {
+		bc = q.baseCache
+	} else {
+		bc = newBaseCache()
+	}
+
 	e := &eval{
 		ctx:                         ctx,
 		metrics:                     q.metrics,
@@ -557,8 +578,7 @@ func (q *Query) Iter(ctx context.Context, iter func(QueryResult) error) error {
 		bindings:                    newBindings(0, q.instr),
 		compiler:                    q.compiler,
 		store:                       q.store,
-		baseCache:                   newBaseCache(),
-		targetStack:                 newRefStack(),
+		baseCache:                   bc,
 		txn:                         q.txn,
 		input:                       q.input,
 		external:                    q.external,
@@ -568,12 +588,10 @@ func (q *Query) Iter(ctx context.Context, iter func(QueryResult) error) error {
 		instr:                       q.instr,
 		builtins:                    q.builtins,
 		builtinCache:                builtins.Cache{},
-		functionMocks:               newFunctionMocksStack(),
 		interQueryBuiltinCache:      q.interQueryBuiltinCache,
 		interQueryBuiltinValueCache: q.interQueryBuiltinValueCache,
 		ndBuiltinCache:              q.ndBuiltinCache,
 		virtualCache:                vc,
-		comprehensionCache:          newComprehensionCache(),
 		genvarprefix:                q.genvarprefix,
 		runtime:                     q.runtime,
 		indexing:                    q.indexing,
